# -*- coding: utf-8 -*-
#
#       media_overlays/base.py
#
#       Copyright 2015 Cimbali <me@cimba.li>
#
#       Vaguely inspired from:
#       gtk example/widget for VLC Python bindings
#       Copyright (C) 2009-2010 the VideoLAN team
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#
"""
:mod:`pympress.media_overlays.base` -- base widget to play videos with an unspecified backend
---------------------------------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, Gio

from pympress import builder


class VideoOverlay(builder.Builder):
    """ Simple Video widget.

    All do_X() functions are meant to be called from the main thread, through e.g. :function:`~GLib.idle_add`,
    for thread-safety in the handling of video backends.

    Args:
        container (:class:`~Gtk.Overlay`): The container with the slide, at the top of which we add the movie area
        page_type (:class:`~pympress.document.PdfPage`): the part of the page to display
        action_map (:class:`~Gio.ActionMap`): the action map that contains the actions for this media
        media (:class:`~pympress.document.Media`): the object defining the properties of the video such as position etc.
    """
    #: :class:`~Gtk.Overlay` that is the parent of the VideoOverlay widget.
    parent = None
    #: :class:`~Gtk.VBox` that contains all the elements to be overlayed.
    media_overlay = None
    #: A :class:`~Gtk.HBox` containing a toolbar with buttons and :attr:`~progress` the progress bar
    toolbar = None
    #: :class:`~Gtk.Scale` that is the progress bar in the controls toolbar - if we have one.
    progress = None
    #: :class:`~Gtk.DrawingArea` where the media is rendered.
    movie_zone = None
    #: `tuple` containing the left/top/right/bottom space around the drawing area in the PDF page
    relative_page_margins = None
    #: `tuple` containing the left/top/right/bottom space around the drawing area in the visible slide
    relative_margins = None
    #: `bool` that tracks whether we should play automatically
    autoplay = False
    #: `bool` that tracks whether we should play after we finished playing
    repeat = False
    #: `str` representing the mime type of the media file
    media_type = ''
    #: `float` giving the initial starting position for playback
    start_pos = 0.
    #: `float` giving the end position for playback, if any
    end_pos = None
    #: `float` representing the last know timestamp at which the progress bar updated
    last_timestamp = 0.

    #: `bool` that tracks whether the user is dragging the position
    dragging_position = False
    #: `bool` that tracks whether the playback was paused when the user started dragging the position
    dragging_paused = False
    #: Format of the video time, defaults to m:ss, changed to m:ss / m:ss when the max time is known
    time_format = '{:01}:{:02}'
    #: `float` holding the max time in s
    maxval = 1

    #: :class:`~Gio.ActionMap` containing the actios for this video overlay
    action_map = None

    def __init__(self, container, page_type, action_map, media):
        super(VideoOverlay, self).__init__()

        self.parent = container
        self.relative_page_margins = tuple(getattr(media.relative_margins, v) for v in ('x1', 'y2', 'x2', 'y1'))
        self.update_margins_for_page(page_type)

        self.load_ui('media_overlay')
        self.toolbar.set_visible(media.show_controls)

        self.connect_signals(self)

        # medias, here the actions are scoped to the current widget
        self.action_map = action_map
        self.media_overlay.insert_action_group('media', self.action_map)
        if media.type:
            self.media_type = media.type
        else:
            content_type, _ = Gio.content_type_guess(media.filename.as_uri())
            self.media_type = Gio.content_type_get_mime_type(content_type)
        self.start_pos = media.start_pos
        self.end_pos = float('inf') if media.duration == 0 else media.start_pos + media.duration
        self._set_file(media.filename)

        self.autoplay = media.autoplay
        self.repeat = media.repeat
        # TODO: handle poster


    def handle_embed(self, mapped_widget):
        """ Handler to embed the video player in the window, connected to the :attr:`~.Gtk.Widget.signals.map` signal.
        """
        return False


    def format_millis(self, sc, prog):
        """ Callback to format the current timestamp (in milliseconds) as minutes:seconds.

        Args:
            sc (:class:`~Gtk.Scale`): The scale whose position we are formatting
            prog (`float`): The position of the :class:`~Gtk.Scale`, i.e. the number of seconds elapsed
        """
        return self.time_format.format(*divmod(int(round(prog)), 60))


    def update_range(self, max_time):
        """ Update the toolbar slider size.

        Args:
            max_time (`float`): The maximum time in this video in s
        """
        self.maxval = max_time
        self.progress.set_range(0, self.maxval)
        self.progress.set_increments(min(5., self.maxval / 10.), min(60., self.maxval / 10.))
        sec = round(self.maxval) if self.maxval > .5 else 1.
        self.time_format = '{{:01}}:{{:02}} / {:01}:{:02}'.format(*divmod(int(sec), 60))


    def update_progress(self, time):
        """ Update the toolbar slider to the current time.

        Args:
            time (`float`): The time in this video in s
        """
        self.progress.set_value(time)


    def progress_moved(self, rng, sc, val):
        """ Callback to update the position of the video when the user moved the progress bar.

        Args:
            rng (:class:`~Gtk.Range`): The range corresponding to the scale whose position we are formatting
            sc (:class:`~Gtk.Scale`): The scale whose position we are updating
            val (`float`): The position of the :class:`~Gtk.Scale`, which is the number of seconds elapsed in the video
        """
        return self.action_map.lookup_action('set_time').activate(GLib.Variant.new_double(val))


    def play_pause(self, *args):
        """ Callback to toggle play/pausing from clicking on the DrawingArea
        """
        return self.action_map.lookup_action('pause').activate()


    def handle_end(self):
        """ End of the stream reached: restart if looping, otherwise hide overlay
        """
        if not self.repeat:
            self.action_map.lookup_action('stop').activate()
        else:
            self.action_map.lookup_action('set_time').activate(GLib.Variant.new_double(self.start_pos))


    def update_margins_for_page(self, page_type):
        """ Recalculate the margins around the media in the event of a page type change.

        Arguments:
            page_type (:class:`~pympress.document.PdfPage`): the part of the page to display
        """
        self.relative_margins = page_type.to_screen(*self.relative_page_margins)


    def resize(self):
        """ Adjust the position and size of the media overlay.
        """
        if not self.is_shown():
            return

        pw, ph = self.parent.get_allocated_width(), self.parent.get_allocated_height()
        self.media_overlay.props.margin_left   = pw * max(self.relative_margins[0], 0)
        self.media_overlay.props.margin_right  = pw * max(self.relative_margins[2], 0)
        self.media_overlay.props.margin_bottom = ph * max(self.relative_margins[3], 0)
        self.media_overlay.props.margin_top    = ph * max(self.relative_margins[1], 0)


    def is_shown(self):
        """ Returns whether the media overlay is currently added to the overlays, or hidden.

        Returns:
            `bool`: `True` iff the overlay is currently displayed.
        """
        return self.media_overlay.get_parent() is not None


    def is_playing(self):
        """ Returns whether the media is currently playing (and not paused).

        Returns:
            `bool`: `True` iff the media is playing.
        """
        raise NotImplementedError


    def do_stop(self):
        """ Stops playing in the backend player.
        """
        raise NotImplementedError


    def _set_file(self, filepath):
        """ Sets the media file to be played by the widget.

        Args:
            filepath (`pathlib.Path`): The path to the media file path
        """
        raise NotImplementedError


    def show(self):
        """ Bring the widget to the top of the overlays if necessary.
        """
        if min(self.relative_margins) < 0:
            logger.warning('Negative margin(s) clipped to 0 (might alter the aspect ratio?): ' +
                           'LTRB = {}'.format(self.relative_margins))

        if not self.media_overlay.get_parent():
            self.parent.add_overlay(self.media_overlay)
            self.parent.reorder_overlay(self.media_overlay, 2)
            self.resize()
            self.parent.queue_draw()
        self.media_overlay.show()


    def do_hide(self, *args):
        """ Remove widget from overlays. Needs to be called via :func:`~GLib.idle_add`.

        Returns:
            `bool`: `True` iff this function should be run again (:func:`~GLib.idle_add` convention)
        """
        self.do_stop()
        self.media_overlay.hide()

        if self.media_overlay.get_parent():
            self.parent.remove(self.media_overlay)
        self.parent.queue_draw()
        return False


    def do_play(self):
        """ Start playing the media file.

        Should run on the main thread to ensure we avoid reentrency problems.

        Returns:
            `bool`: `True` iff this function should be run again (:meth:`~GLib.idle_add` convention)
        """
        raise NotImplementedError


    def do_play_pause(self):
        """ Toggle pause mode of the media.

        Should run on the main thread to ensure we avoid reentrency problems.

        Returns:
            `bool`: `True` iff this function should be run again (:meth:`~GLib.idle_add` convention)
        """
        raise NotImplementedError


    def do_set_time(self, t):
        """ Set the player at time t.

        Should run on the main thread to ensure we avoid vlc plugins' reentrency problems.

        Args:
            t (`float`): the timestamp, in s

        Returns:
            `bool`: `True` iff this function should be run again (:meth:`~GLib.idle_add` convention)
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
#
#       media_overlays/vlc.py
#
#       Copyright 2018 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#
"""
:mod:`pympress.media_overlays.vlc` -- widget to play videos using VLC
---------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import os
import vlc
import ctypes

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib

from pympress.util import IS_WINDOWS
from pympress.media_overlays import base


def get_window_handle(window):
    """ Uses ctypes to call gdk_win32_window_get_handle which is not available in python gobject introspection porting.

    Solution from http://stackoverflow.com/a/27236258/1387346

    Args:
        window (:class:`~Gdk.Window`): The window for which we want to get the handle

    Returns:
        The handle to the win32 window
    """
    # get the c gpointer of the gdk window
    ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_void_p
    ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object]
    drawingarea_gpointer = ctypes.pythonapi.PyCapsule_GetPointer(window.__gpointer__, None)
    # get the win32 handle
    gdkdll = ctypes.CDLL('libgdk-3-0.dll')
    handle_getter = gdkdll.gdk_win32_window_get_handle
    handle_getter.restype = ctypes.c_void_p
    handle_getter.argtypes = [ctypes.c_void_p]
    return handle_getter(drawingarea_gpointer)


class VlcOverlay(base.VideoOverlay):
    """ Simple VLC widget.

    Its player can be controlled through the 'player' attribute, which is a :class:`~vlc.MediaPlayer` instance.
    """

    #: A single vlc.Instance() to be shared by (possible) multiple players.
    _instance = None

    def __init__(self, *args, **kwargs):
        self.player = self._instance.media_player_new()  # before loading UI, needed to connect "map" signal

        super(VlcOverlay, self).__init__(*args, **kwargs)
        # Simple black background painting to avoid glitching outside of video area
        if not self.media_type.startswith('audio'):
            self.movie_zone.connect('draw', self.paint_backdrop)

        event_manager = self.player.event_manager()
        event_manager.event_attach(vlc.EventType.MediaPlayerEndReached, lambda e: GLib.idle_add(self.handle_end))
        event_manager.event_attach(vlc.EventType.MediaPlayerLengthChanged,
                                   lambda e: self.update_range(self.player.get_length() / 1000. or 1.))
        event_manager.event_attach(vlc.EventType.MediaPlayerTimeChanged, self.time_changed)


    def handle_embed(self, mapped_widget):
        """ Handler to embed the VLC player in the window, connected to the :attr:`~.Gtk.Widget.signals.map` signal.
        """
        # Do we need to be on the main thread? (especially for the mess from the win32 window handle)
        # assert(isinstance(threading.current_thread(), threading._MainThread))
        window = self.movie_zone.get_window()
        if window is None:
            logger.error('No window in which to embed the VLC player!')
            return False
        elif IS_WINDOWS:
            self.player.set_hwnd(get_window_handle(window))  # get_property('window')
        else:
            self.player.set_xwindow(window.get_xid())
        self.movie_zone.queue_draw()
        return False


    def is_playing(self):
        """ Returns whether the media is currently playing (and not paused).

        Returns:
            `bool`: `True` iff the media is playing.
        """
        return self.player.is_playing()


    def _set_file(self, filepath):
        """ Sets the media file to be played by the widget.

        Args:
            filepath (`pathlib.Path`): The path to the media file path
        """
        self.player.set_media(self._instance.media_new(filepath.resolve().as_uri()))


    def handle_end(self):
        """ End of the stream reached: restart if looping, otherwise hide overlay

        Overrided because, to implement looping, vlc plugin needs to be told to start on stream end, not to seek
        """
        if self.repeat:
            self.action_map.lookup_action('play').activate()
        else:
            self.action_map.lookup_action('stop').activate()


    def mute(self, value):
        """ Mutes the player.

        Args:
            value (`bool`): `True` iff this player should be muted
        """
        GLib.idle_add(self.player.audio_set_volume, 0 if value else 100)
        return False


    def do_play(self):
        """ Start playing the media file.

        Should run on the main thread to ensure we avoid vlc plugins' reentrency problems.

        Returns:
            `bool`: `True` iff this function should be run again (:func:`~GLib.idle_add` convention)
        """
        play_from_state = self.player.get_state()
        if play_from_state in {vlc.State.Ended, vlc.State.Playing}:
            self.player.stop()
            play_from_state = vlc.State.Stopped

        self.player.play()

        if play_from_state in {vlc.State.NothingSpecial, vlc.State.Stopped}:
            self.do_set_time(self.start_pos)

        self.movie_zone.queue_draw()
        return False


    def paint_backdrop(self, widget, context):
        """ Draw behind/around the video, aka the black bars

        Args:
            widget (:class:`~Gtk.Widget`):  the widget to update
            context (:class:`~cairo.Context`):  the Cairo context (or `None` if called directly)
        """
        context.save()
        context.set_source_rgb(0, 0, 0)
        context.fill()
        context.paint()
        context.restore()


    def show(self):
        """ Bring the widget to the top of the overlays if necessary − also force redraw of movie zone
        """
        super(VlcOverlay, self).show()
        self.movie_zone.queue_draw()


    def do_play_pause(self):
        """ Toggle pause mode of the media.

        Should run on the main thread to ensure we avoid vlc plugins' reentrency problems.

        Returns:
            `bool`: `True` iff this function should be run again (:func:`~GLib.idle_add` convention)
        """
        self.player.pause() if self.player.is_playing() else self.player.play()
        return False


    def do_stop(self):
        """ Stops playing in the backend player.
        """
        self.player.stop()


    def time_changed(self, event):
        """ Handle time passing

        Args:
            event (:class:`~vlc.Event`): The event that triggered the handler
        """
        time = self.player.get_time() / 1000. or 1.
        if self.last_timestamp <= self.end_pos <= time:
            self.handle_end()
        self.last_timestamp = time
        self.update_progress(time)


    def do_set_time(self, time):
        """ Set the player at time `~time`.

        Should run on the main thread to ensure we avoid vlc plugins' reentrency problems.

        Args:
            `~time` (`float`): the timestamp, in s

        Returns:
            `bool`: `True` iff this function should be run again (:func:`~GLib.idle_add` convention)
        """
        # Update last_timestamp first, as seeking should bypass auto stop after duration
        self.last_timestamp = time
        self.player.set_time(int(round(time * 1000.)))
        return False


    @classmethod
    def setup_backend(cls, vlc_opts = ['--no-video-title-show']):
        """ Prepare/check the VLC backend.

        Args:
            vlc_opts (`list`): the arguments for starting vlc

        Returns:
            `str`: the version of VLC used by the backend
        """
        if IS_WINDOWS and vlc.plugin_path:
            # let python find the DLLs
            os.environ['PATH'] = vlc.plugin_path + ';' + os.environ['PATH']

        VlcOverlay._instance = vlc.Instance(vlc_opts)
        return 'VLC {}'.format(vlc.libvlc_get_version().decode('ascii'))



# -*- coding: utf-8 -*-
#
#       deck.py
#
#       Copyright 2023 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.deck` -- Manage user drawings on the current slide
---------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib

from pympress import builder


class Overview(builder.Builder):
    """ UI that allows to draw free-hand on top of the current slide.

    Args:
        config (:class:`~pympress.config.Config`): A config object containing preferences
        builder (:class:`~pympress.builder.Builder`): A builder from which to load widgets
        notes_mode (`bool`): The current notes mode, i.e. whether we display the notes on second slide
    """
    #: Whether we are displaying the deck overview on screen
    deck_mode = False

    #: :class:`~Gtk.Viewport` that replaces normal panes when deck is shown
    deck_viewport = None
    #: :class:`~Gtk.Grid` that displays all the slides of the overview
    deck_grid = None
    #: The :class:`~Gtk.DrawingArea` for the first slide
    deck0 = None

    #: A :class:`~Gtk.OffscreenWindow` where we render the deck interface when it's not shown
    deck_off_render = None
    #: :class:`~Gtk.Box` in the Presenter window, where we insert deck.
    p_central = None

    #: callback, to be connected to :func:`~pympress.ui.UI.compute_frame_grid`
    compute_frame_grid = lambda *args: None

    #: callback, to be connected to :func:`~pympress.ui.UI.load_layout`
    load_layout = lambda *args: None

    #: callback, to be connected to :func:`~pympress.surfacecache.SurfaceCache.resize_widget`
    resize_cache = lambda *args: None
    #: callback, to be connected to :func:`~pympress.ui.UI.goto_page`
    goto_page = lambda *args: None
    #: :class:`~pympress.surfacecache.SurfaceCache` instance.
    cache = None

    #: `tuple` of rows/columns in the grid
    grid_size = (0, 0)
    #: `bool` whether we show all pages or remove consecutive identically labeled pages, keeping only the last
    all_pages = False
    #: `int` How large (at most) to make rows
    max_row_size = 6

    #: The :class:`~Gtk.DrawingArea` in the content window
    c_da = None

    def __init__(self, config, builder, notes_mode):
        super(Overview, self).__init__()

        self.cache = builder.cache
        self.load_ui('deck')
        builder.load_widgets(self)
        self.deck_da_list = [self.deck0]
        self.get_application().add_window(self.deck_off_render)

        self.load_layout = builder.get_callback_handler('load_layout')
        self.goto_page = builder.get_callback_handler('goto_page')
        self.compute_frame_grid = builder.get_callback_handler('compute_frame_grid')
        self.setup_doc_callbacks(builder.doc)

        self.connect_signals(self)

        self.max_row_size = config.getint('deck-overview', 'max-slides-per-row')
        # Whether to show all pages or only distinctly labeled pages (useful for latex)
        self.all_pages = not config.get('deck-overview', 'distinct-labels-only')

        self.setup_actions({
            'deck-overview': dict(activate=self.switch_deck_overview, state=False),
        })


    def on_deck_hover(self, widget, event):
        """ Track when each deck in the slide is hovered
        """
        ctx = widget.get_style_context()
        ctx.set_state(Gtk.StateFlags.PRELIGHT if event.type == Gdk.EventType.ENTER_NOTIFY else Gtk.StateFlags.NORMAL)
        widget.queue_draw()


    def setup_doc_callbacks(self, doc):
        """ Callbacks that need to be setup again at every new document

        Args:
            doc (:class:`~pympress.document.Document`): The new document that got loaded
        """
        self.pages_number         = doc.pages_number
        self.has_labels           = doc.has_labels
        self.get_last_label_pages = doc.get_last_label_pages

        self.create_drawing_areas()


    def try_cancel(self):
        """ Cancel deck, if it is enabled.

        Returns:
            `bool`: `True` if deck got cancelled, `False` if it was already disabled.
        """
        if not self.deck_mode:
            return False

        self.disable_deck_overview()
        return True


    def create_drawing_areas(self):
        """ Build DrawingArea and AspectFrame elements to display later on
        """
        pages = self.get_last_label_pages() if self.has_labels() else range(self.pages_number())
        self.grid_size = (len(pages), 1)

        # Always keep the first drawing area as it is used to provide surfaces in the cache
        for row in range(1, self.grid_size[0]):
            self.deck_grid.remove_row(row)
        for col in range(1, self.grid_size[1]):
            self.deck_grid.remove_row(col)

        # Set drawing areas
        for num, da in enumerate(self.deck_da_list[1:], 1):
            self.deck_grid.attach(da.get_parent(), 1, num, 1, 1)

        for num in range(len(self.deck_da_list), len(pages)):
            da = Gtk.DrawingArea()
            da.add_events(Gdk.EventMask.TOUCH_MASK | Gdk.EventMask.ENTER_NOTIFY_MASK | Gdk.EventMask.LEAVE_NOTIFY_MASK |
                          Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.BUTTON_RELEASE_MASK)
            da.connect('draw', self.on_deck_draw)
            da.connect('button-release-event', self.on_deck_click)
            da.connect('touch-event', self.on_deck_click)
            da.connect('enter-notify-event', self.on_deck_hover)
            da.connect('leave-notify-event', self.on_deck_hover)
            self.deck_da_list.append(da)

            frame = Gtk.AspectFrame()
            frame.get_style_context().add_class('grid-frame')
            frame.set_shadow_type(Gtk.ShadowType.NONE)
            frame.add(da)

            self.deck_grid.attach(frame, 1, num, 1, 1)

        ratio = self.c_da.get_allocated_width() / self.c_da.get_allocated_height()
        for page, da in zip(pages, self.deck_da_list):
            da.get_parent().set(.5, .5, ratio, False)
            da.set_name('deck{}'.format(page))


    def reset_grid(self, *args):
        """ Set the slides configuration and size in the grid
        """
        # Gather info about slides to display
        num_pages = self.pages_number() if self.all_pages or not self.has_labels() else len(self.get_last_label_pages())
        ratio = self.c_da.get_allocated_width() / self.c_da.get_allocated_height()

        ww, wh = self.deck_grid.get_allocated_width(), self.deck_grid.get_allocated_height()
        sw, sh = self.deck_grid.get_row_spacing(), self.deck_grid.get_column_spacing()
        rows, cols = self.compute_frame_grid(ww / wh, num_pages)
        window = self.deck_viewport.get_window()
        scale = window.get_scale_factor()

        if not rows or not cols:
            rows, cols = 1, 1

        dw, dh = (ww + sw) / cols - sw, (wh + sh) / rows - sh
        dw, dh = min(dw, dh * ratio), min(dw / ratio, dh)

        if cols > self.max_row_size:
            cols = self.max_row_size
            rows = (num_pages + cols - 1) // cols

            dw = (ww + sw) / cols - sw
            dh = dw / ratio

        self.cache.resize_widget('deck', int(dw * scale), int(dh * scale))

        frames = [da.get_parent() for da in self.deck_da_list]
        for frame in frames:
            frame.set(.5, .5, ratio, False)
            frame.set_size_request(dw / scale, dh / scale)

        if self.grid_size != (rows, cols):
            for frame in frames[1:]:
                self.deck_grid.remove(frame)

            # Always keep the first drawing area as it is used to provide surfaces in the cache
            for row in range(rows, self.grid_size[0]):
                self.deck_grid.remove_row(row)
            for col in range(cols, self.grid_size[1]):
                self.deck_grid.remove_row(col)

            for num, frame in enumerate(frames[1:], 1):
                self.deck_grid.attach(frame, num % cols, num // cols, 1, 1)

            # resize grid and cache
            self.grid_size = rows, cols

        for da in self.deck_da_list:
            GLib.idle_add(self.prerender, da)

        self.deck_grid.show_all()


    def prerender(self, da):
        """ Perform in-cache rendering

        Args:
            da (:class:`~Gtk.DrawingArea`):  the widget for which we’re rendering
        """
        self.cache.renderer('deck', int(da.get_name()[4:]))
        da.queue_draw()
        return GLib.SOURCE_REMOVE


    def on_deck_draw(self, widget, cairo_context):
        """ Actually draw the deck slide -- only do this from cache, to limit overhead

        Args:
            widget (:class:`~Gtk.Widget`):  the widget to update
            cairo_context (:class:`~cairo.Context`):  the Cairo context (or `None` if called directly)
        """
        page_num = int(widget.get_name()[4:])
        pb = self.cache.get('deck', page_num)
        if pb is None:
            # We’ll redraw later
            widget.queue_draw()
            return

        window = widget.get_window()
        scale = window.get_scale_factor()
        cairo_context.scale(1. / scale, 1. / scale)
        cairo_context.set_source_surface(pb, 0, 0)
        cairo_context.paint()

        ctx = widget.get_style_context()
        if ctx.get_state() != Gtk.StateFlags.PRELIGHT:
            return

        # Draw a hover border manually
        color = ctx.get_property('border-color', Gtk.StateFlags.PRELIGHT)
        width = 2
        cairo_context.set_source_rgba(*color)
        cairo_context.set_line_width(width)

        ww, wh = widget.get_allocated_width(), widget.get_allocated_height()
        cairo_context.move_to(width / 2, width / 2)
        cairo_context.line_to(width / 2, wh - width / 2)
        cairo_context.line_to(ww - width / 2, wh - width / 2)
        cairo_context.line_to(ww - width / 2, width / 2)
        cairo_context.close_path()
        cairo_context.stroke()


    def on_deck_click(self, widget, event):
        """ A slide has been clicked, go to it

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the key stroke
            event (:class:`~Gdk.Event`):  the GTK event, which contains the key stroke details
        """
        page_num = int(widget.get_name()[4:])
        self.goto_page(page_num, False)
        self.disable_deck_overview()


    def switch_deck_overview(self, gaction, target=None):
        """ Starts the mode where one can read on top of the screen.

        Args:

        Returns:
            `bool`: whether the event was consumed
        """
        if target is not None and target == self.deck_mode:
            return False

        # Perform the state toggle
        if self.deck_mode:
            return self.disable_deck_overview()
        else:
            return self.enable_deck_overview()


    def enable_deck_overview(self):
        """ Enable the deck view.

        Returns:
            `bool`: whether it was possible to enable (thus if it was not enabled already)
        """
        if self.deck_mode:
            return False

        self.deck_off_render.remove(self.deck_viewport)
        self.load_layout('deck-overview')

        self.p_central.queue_draw()
        self.deck_viewport.queue_draw()

        GLib.idle_add(self.reset_grid)

        self.deck_mode = True
        self.get_application().lookup_action('deck-overview').change_state(GLib.Variant.new_boolean(self.deck_mode))

        self.p_central.queue_draw()
        return True


    def disable_deck_overview(self):
        """ Disable the deck view.

        Returns:
            `bool`: whether it was possible to disable (thus if it was not disabled already)
        """
        if not self.deck_mode:
            return False

        self.deck_mode = False

        self.load_layout(None)
        self.deck_off_render.add(self.deck_viewport)

        self.get_application().lookup_action('deck-overview').change_state(GLib.Variant.new_boolean(self.deck_mode))

        self.p_central.queue_draw()
        return True



# -*- coding: utf-8 -*-
#
#       extras.py
#
#       Copyright 2017 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.extras` -- Manages the display of fancy extras such as annotations, videos and cursors
-----------------------------------------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import pathlib
import mimetypes
import functools
from urllib.request import url2pathname

import gi
import cairo
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib, Gio

from pympress import document, builder, util


class Annotations(object):
    """ Widget displaying a PDF’s text annotations.
    """
    #: The containing :class:`~Gtk.TreeView` widget for the annotations
    annotations_treeview = None
    #: The containing :class:`~Gtk.ListStore` storing the annotations to be displayed
    annotations_liststore = None
    #: The :class:`~Gtk.Entry` in which we are currently editing an annotation, or None
    editing = None

    new_doc_annotation = lambda *args: None
    set_doc_annotation = lambda *args: None
    remove_doc_annotation = lambda *args: None

    def __init__(self, builder):
        super(Annotations, self).__init__()
        builder.load_widgets(self)
        builder.setup_actions({
            'add-annotation': dict(activate=self.add_annotation),
            'remove-annotation': dict(activate=self.remove_annotation),
        })


    def load_annotations(self, annot_page):
        """ Add annotations to be displayed (typically on going to a new slide).

        Args:
            annot_page (:class:`~pympress.document.Page`): The page object that contains the annotations
        """
        self.annotations_liststore.clear()
        for num, annot in enumerate(annot_page.get_annotations()):
            self.annotations_liststore.append([annot.get_contents(), num])

        self.new_doc_annotation = annot_page.new_annotation
        self.set_doc_annotation = annot_page.set_annotation
        self.remove_doc_annotation = annot_page.remove_annotation


    def try_cancel(self):
        """ Try to cancel editing

        Returns:
            `bool`: whether editing was enabled and thus canceled
        """
        if self.editing is None:
            return False

        rows = self.annotations_treeview.get_selection().get_selected_rows()[1]
        if rows:
            self.annotations_treeview.set_cursor(rows[0], None, False)
        return True


    def key_event(self, widget, event):
        """ Handle a key (press/release) event.

        Needed to forward events directly to the :class:`~Gtk.Entry`, bypassing the global action accelerators.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.

        Returns:
            `bool`: whether the event was consumed
        """
        if self.editing is None:
            return False
        elif event.get_event_type() == Gdk.EventType.KEY_PRESS:
            return self.editing.do_key_press_event(self.editing, event)
        elif event.get_event_type() == Gdk.EventType.KEY_RELEASE:
            return self.editing.do_key_release_event(self.editing, event)
        return False


    def editing_started(self, cell_renderer, widget, entry_number):
        """ Handle edit start

        Args:
            cell_renderer (:class:`~Gtk.CellRenderer`): The renderer which received the signal
            widget (:class:`~Gtk.CellEditable`): the Gtk entry editing the annotation entry
            entry_number (`str`): the string representation of the path identifying the edited cell
        """
        self.editing = widget


    def editing_validated(self, cell_renderer, entry_number, new_content):
        """ Handle successful edit: store the new cell value in the model and the document

        Args:
            cell_renderer (:class:`~Gtk.CellRenderer`): The renderer which received the signal
            entry_number (`str`): the string representation of the path identifying the edited cell
            new_content (`str`): the new value of the edited cell
        """
        path = Gtk.TreePath.new_from_string(entry_number)
        row = self.annotations_liststore.get_iter(path)
        self.annotations_liststore.set_value(row, 0, new_content)
        self.set_doc_annotation(path.get_indices()[0], new_content)
        self.editing_finished(cell_renderer)


    def editing_finished(self, cell_renderer):
        """ Handle the end of editing

        Args:
            cell_renderer (:class:`~Gtk.CellRenderer`): The renderer which received the signal
        """
        self.editing = None


    def add_annotation(self, gaction, param=None):
        """ Add an annotation to the the page’s annotation list

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call, which identifies which backend
            param (:class:`~GLib.Variant`): an optional parameter
        """
        path = self.annotations_liststore.get_path(self.annotations_liststore.append())
        self.annotations_treeview.set_cursor(path, self.annotations_treeview.get_columns()[0], True)
        self.new_doc_annotation(path.get_indices()[0])
        return True


    def remove_annotation(self, gaction, param=None):
        """ Remove an annotation to the from the page’s annotation list

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call, which identifies which backend
            param (:class:`~GLib.Variant`): an optional parameter
        """
        rows = self.annotations_treeview.get_selection().get_selected_rows()[1]
        if not rows:
            return False
        self.annotations_liststore.remove(self.annotations_liststore.get_iter(rows[0]))
        self.remove_doc_annotation(rows[0].get_indices()[0])
        return True



class Media(object):
    """ Class managing statically the medias and media player backends, to enable play/pause callbacks.

    Args:
        builder (:class:`~pympress.builder.Builder`): A builder from which to load widgets
        conf (:class:`~pympress.config.Config`): An object containing the preferences
    """
    #: `dict` of :class:`~pympress.media_overlays.base.VideoOverlay` ready to be added on top of the slides
    _media_overlays = {}

    #: :class:`~Gtk.Overlay` for the Content window.
    c_overlay = None
    #: :class:`~Gtk.Overlay` for the Presenter window.
    p_overlay = None

    #: `dict` with the backend modules that were correctly loaded, mapping backend identifiers (`str`)
    #: to :class:`~pympress.media_overlays.base.VideoOverlay` sub-classes
    _backends = {}
    #: `dict` containing backends and their mappings to mime type lists for which they are enabled.
    #: A default backend is marked by an empty list.
    types_list = {}

    def __init__(self, builder, conf):
        super(Media, self).__init__()
        self.conf = conf

        self._setup_backends()
        builder.load_widgets(self)

        builder.setup_actions({
            'use-{}-backend'.format(backend): {
                'activate': self.toggle,
                'state': backend in self.types_list,
                'enabled': backend in self._backends,
            } for backend in self._backends
        })

        self.c_overlay.queue_draw()
        self.p_overlay.queue_draw()


    def toggle(self, gaction, param=None):
        """ Toggle a backend (if it was loaded correctly)

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call, which identifies which backend
            param (:class:`~GLib.Variant`): an optional parameter
        """
        backend = gaction.get_name().split('-')[1]
        if backend not in self._backends:
            return ValueError('Unexpected backend')

        enable = backend not in self.types_list

        if enable:
            self.types_list[backend] = self.conf.getlist(backend, 'mime_types')
        else:
            del self.types_list[backend]

        gaction.change_state(GLib.Variant.new_boolean(enable))
        self.conf.set(backend, 'enabled', 'on' if enable else 'off')


    def remove_media_overlays(self):
        """ Remove current media overlays.
        """
        for media_id in self._media_overlays:
            self.hide(media_id)


    def purge_media_overlays(self):
        """ Remove current media overlays.
        """
        self.remove_media_overlays()
        self._media_overlays.clear()


    def replace_media_overlays(self, current_page, page_type):
        """ Remove current media overlays, add new ones if page contains media.

        Args:
            current_page (:class:`~pympress.document.Page`): The page for which to prepare medias
            page_type (:class:`~pympress.document.PdfPage`): The part of the page to consider
        """
        if page_type == document.PdfPage.NONE:
            return

        self.remove_media_overlays()

        for media in current_page.get_media():
            media_id = hash(media)

            if media_id not in self._media_overlays:
                mime_type, enc = mimetypes.guess_type(media.filename)
                factory = self.get_factory(mime_type)

                if not factory:
                    logger.warning('No available overlay for mime type {}, ignoring media {}'
                                   .format(mime_type, media.filename))
                    continue

                action_group = Gio.SimpleActionGroup.new()
                builder.Builder.setup_actions({
                    'play':     dict(activate=functools.partial(self.play, media_id)),
                    'stop':     dict(activate=functools.partial(self.hide, media_id)),
                    'pause':    dict(activate=functools.partial(self.play_pause, media_id)),
                    'set_time': dict(activate=functools.partial(self.set_time, media_id), parameter_type=float)
                }, action_group)

                v_da_c = factory(self.c_overlay, page_type, action_group, media)
                v_da_p = factory(self.p_overlay, page_type, action_group, media._replace(show_controls=True))

                self._media_overlays[media_id] = (v_da_c, v_da_p)

            self._media_overlays[media_id][0].mute(True)
            self._media_overlays[media_id][1].mute(False)

            if any(overlay.autoplay for overlay in self._media_overlays[media_id]):
                GLib.idle_add(self.play, media_id)
            # TODO: handle poster


    def resize(self, which=None):
        """ Resize all media overlays that are a child of an overlay.
        """
        needs_resizing = (which == 'content', which == 'presenter') if which is not None else (True, True)
        for media_id in self._media_overlays:
            for widget in (w for w, r in zip(self._media_overlays[media_id], needs_resizing) if r and w.is_shown()):
                widget.resize()


    def adjust_margins_for_mode(self, page_type):
        """ Adjust the relative margins of child widgets for notes mode update.

        Args:
            page_type (:class:`~pympress.document.PdfPage`): The part of the page to display
        """
        for media_id in self._media_overlays:
            for widget in self._media_overlays[media_id]:
                widget.update_margins_for_page(page_type)


    def play(self, media_id, gaction=None, param=None):
        """ Starts playing a media. Used as a callback.

        Args:
            media_id (`int`): A unique identifier of the media to start playing
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        if media_id in self._media_overlays:
            c, p = self._media_overlays[media_id]
            p.show()
            c.show()
            GLib.idle_add(lambda: any(p.do_play() for p in self._media_overlays[media_id]))


    def hide(self, media_id, gaction=None, param=None):
        """ Stops playing a media and hides the player. Used as a callback.

        Args:
            media_id (`int`): A unique identifier of the media to start playing
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        if media_id in self._media_overlays:
            c, p = self._media_overlays[media_id]
            if c.is_shown(): GLib.idle_add(c.do_hide)
            if p.is_shown(): GLib.idle_add(p.do_hide)


    def hide_all(self):
        """ Stops all playing medias and hides the players. Used before exit.
        """
        for c, p in self._media_overlays.values():
            if c.is_shown(): GLib.idle_add(c.do_hide)
            if p.is_shown(): GLib.idle_add(p.do_hide)


    def play_pause(self, media_id, gaction=None, param=None):
        """ Toggles playing and pausing a media. Used as a callback.

        Args:
            media_id (`int`): A unique idientifier of the media to start playing
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        GLib.idle_add(lambda: any(p.do_play_pause() for p in self._media_overlays[media_id]))


    def set_time(self, media_id, gaction=None, param=None):
        """ Set the player of a given media at time t. Used as a callback.

        Args:
            media_id (`int`): A unique idientifier of the media to start playing
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): A wrapped float containing the time to which we have to go.
        """
        t = param.get_double()
        GLib.idle_add(lambda: any(p.do_set_time(t) for p in self._media_overlays[media_id]))


    def _setup_backends(self):
        """ Load the backends for video overlays.
        """
        try:
            from pympress.media_overlays.gif_backend import GifOverlay

            gif_version = GifOverlay.setup_backend()
            self._backends['gif'] = GifOverlay
            self.types_list['gif'] = ['image/gif', 'image/svg+xml']

        except Exception as e:
            gif_version = 'GdkPixbuf not available'
            logger.error(_('Media support using {} is disabled.').format('GdkPixbuf'))
            logger.info(_('Caused by ') + type(e).__name__ + ': ' + str(e))


        if not self.conf.getboolean('gstreamer', 'enabled'):
            gst_version = 'GStreamer disabled'
        else:
            try:
                from pympress.media_overlays.gst_backend import GstOverlay
                gst_version = GstOverlay.setup_backend(self.conf.getlist('gstreamer', 'init_options'))
            except Exception as e:
                gst_version = 'GStreamer not available'
                logger.debug(_('Media support using {} is disabled.').format('GStreamer'))
                logger.debug(_('Caused by ') + type(e).__name__ + ': ' + str(e))
            else:
                self._backends['gstreamer'] = GstOverlay
                self.types_list['gstreamer'] = self.conf.getlist('gstreamer', 'mime_types')


        if not self.conf.getboolean('vlc', 'enabled'):
            vlc_version = 'VLC disabled'
        else:
            try:
                from pympress.media_overlays.vlc_backend import VlcOverlay
                vlc_version = VlcOverlay.setup_backend(self.conf.getlist('vlc', 'init_options'))
            except Exception as e:
                vlc_version = 'VLC not available'
                logger.debug(_('Media support using {} is disabled.').format('VLC'))
                logger.debug(_('Caused by ') + type(e).__name__ + ': ' + str(e))
            else:
                self._backends['vlc'] = VlcOverlay
                self.types_list['vlc'] = self.conf.getlist('vlc', 'mime_types')

        self.backend_version = ', '.join([gif_version, gst_version, vlc_version])
        logger.info(_('Media support: ') + self.backend_version)


    def get_factory(self, mime_type):
        """ Returns a class of type :attr:`~_backend`.
        """
        # Search for specific mime type, if no matches fall back to empty lists which mean any mime types
        options = [backend for backend, mime_types in self.types_list.items() if mime_type in mime_types]
        if not options:
            options = [backend for backend, mime_types in self.types_list.items() if len(mime_types) == 0]

        if not options:
            return None

        # Prefer more stable backends with less external dependencies
        priority = ['gif', 'vlc', 'gstreamer'] if util.IS_WINDOWS else ['gif', 'gstreamer', 'vlc']
        return self._backends[sorted(options, key=priority.index)[0]]



class Cursor(object):
    """ Class managing cursors statically for displays, so we can select the mouse cursor with a simple string.
    """
    #: a static `dict` of :class:`~Gdk.Cursor`s, ready to use
    _cursors = {
        'parent': None,
    }

    @classmethod
    def _populate_cursors(cls):
        cls._cursors.update({
            'default': Gdk.Cursor.new_for_display(Gdk.Display.get_default(), Gdk.CursorType.LEFT_PTR),
            'pointer': Gdk.Cursor.new_for_display(Gdk.Display.get_default(), Gdk.CursorType.HAND1),
            'crosshair': Gdk.Cursor.new_for_display(Gdk.Display.get_default(), Gdk.CursorType.CROSSHAIR),
            'invisible': Gdk.Cursor.new_for_display(Gdk.Display.get_default(), Gdk.CursorType.BLANK_CURSOR),
        })

    @classmethod
    def set_cursor(cls, widget, cursor_name = 'parent'):
        """ Set the cursor named cursor_name'.

        Args:
            widget (:class:`~Gtk.Widget`): The widget triggering the cursor change, used to retrieve a Gdk.Window
            cursor_name (`str`): Name of the cursor to be set
        """
        try:
            cursor = cls._cursors[cursor_name]
        except KeyError:
            cls._populate_cursors()
            cursor = cls._cursors[cursor_name]

        window = widget.get_window()
        if window is not None:
            window.set_cursor(cursor)


class Zoom(object):
    """ Manage the zoom level (using a cairo matrix), draw area that will be zoomed while it is being selected.

    Args:
        builder (:class:`~pympress.builder.Builder`): A builder from which to load widgets
    """
    #: Whether we are displaying the interface to scribble on screen and the overlays containing said scribbles
    zoom_selecting = False
    zoom_points = None
    scale = 1.
    shift = (0, 0)

    #: :class:`~Gtk.Box` in the Presenter window, used to reliably set cursors.
    p_central = None
    #: callback, to be connected to :meth:`~pympress.app.Pympress.set_action_enabled`
    set_action_enabled = None

    #: callback, to be connected to :func:`~pympress.ui.UI.redraw_current_slide`
    redraw_current_slide = lambda *args: None
    #: callback, to be connected to :func:`~pympress.ui.UI.clear_cache`
    clear_cache = lambda *args: None

    def __init__(self, builder):
        super(Zoom, self).__init__()
        builder.load_widgets(self)

        self.redraw_current_slide = builder.get_callback_handler('redraw_current_slide')
        self.clear_cache = builder.get_callback_handler('clear_zoom_cache')
        self.set_action_enabled = builder.get_callback_handler('app.set_action_enabled')

        builder.setup_actions({
            'zoom':   dict(activate=self.start_zooming),
            'unzoom': dict(activate=self.stop_zooming),
        })


    def start_zooming(self, *args):
        """ Setup for the user to select the zooming area.

        Returns:
            `bool`: whether the event was consumed
        """
        self.zoom_selecting = True
        Cursor.set_cursor(self.p_central, 'crosshair')

        return True


    def stop_zooming(self, *args):
        """ Cancel the zooming, reset the zoom level to full page.

        Returns:
            `bool`: whether the event was consumed
        """
        Cursor.set_cursor(self.p_central)
        self.zoom_selecting = False
        self.zoom_points = None
        self.scale = 1.
        self.shift = (0, 0)
        self.set_action_enabled('unzoom', False)

        self.redraw_current_slide()
        self.clear_cache()

        return True


    def try_cancel(self):
        """ Cancel the zoom selection, if it was enabled.

        Returns:
            `bool`: `True` if the zoom was cancelled, `False` if a zoom selection was not in progress.
        """
        if not self.zoom_selecting:
            return False

        Cursor.set_cursor(self.p_central)
        self.zoom_selecting = False
        self.zoom_points = None
        return True


    def get_slide_point(self, widget, event):
        """ Gets the point on the slide on a scale (0..1, 0..1), from its position in the widget.
        """
        ww, wh = widget.get_allocated_width(), widget.get_allocated_height()
        ex, ey = event.get_coords()

        return ((ex / ww - self.shift[0]) / self.scale, (ey / wh - self.shift[1]) / self.scale)


    def get_matrix(self, ww, wh):
        """ Returns the :class:`~cairo.Matrix` used to perform the zoom for the widget of size ww x wh.

        Args:
            ww (`float`):  widget width
            wh (`float`):  widget height

        Returns:
            :class:`~cairo.Matrix`: the zoom transformation matrix
        """
        return cairo.Matrix(xx = self.scale, x0 = ww * self.shift[0],
                            yy = self.scale, y0 = wh * self.shift[1])


    def track_zoom_target(self, widget, event):
        """ Draw the zoom's target rectangle.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.

        Returns:
            `bool`: whether the event was consumed
        """
        if self.zoom_selecting and self.zoom_points:
            self.zoom_points[1] = self.get_slide_point(widget, event)

            self.redraw_current_slide()
            return True

        return False


    def toggle_zoom_target(self, widget, event):
        """ Start/stop drawing the zoom's target rectangle.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.

        Returns:
            `bool`: whether the event was consumed
        """
        if not self.zoom_selecting:
            return False

        if event.get_event_type() == Gdk.EventType.BUTTON_PRESS:
            p = self.get_slide_point(widget, event)
            self.zoom_points = [p, p]

            return self.track_zoom_target(widget, event)

        elif event.get_event_type() == Gdk.EventType.BUTTON_RELEASE and self.zoom_points:
            self.zoom_points[1] = self.get_slide_point(widget, event)

            xmin, xmax = sorted(p[0] for p in self.zoom_points)
            ymin, ymax = sorted(p[1] for p in self.zoom_points)
            self.zoom_points = None

            try:
                # zoom by dimension less zoomed, to fit box while maintaining aspect ratio
                self.scale = 1. / max(ymax - ymin, xmax - xmin)

                # make center of drawn rectangle the center of the zoomed slide
                self.shift = (.5 - self.scale * (xmin + xmax) / 2,
                              .5 - self.scale * (ymin + ymax) / 2)
            except ZeroDivisionError:
                self.scale = 1.
                self.shift = (0, 0)

            # stop drawing rectangles and reset cursor (NB don't use window, this bugs)
            Cursor.set_cursor(self.p_central)

            self.zoom_selecting = False
            self.clear_cache()
            self.redraw_current_slide()
            self.set_action_enabled('unzoom', True)

            return True

        return False


    def draw_zoom_target(self, widget, cairo_context):
        """ Perform the drawings by user.

        Args:
            widget (:class:`~Gtk.DrawingArea`): The widget where to draw the scribbles.
            cairo_context (:class:`~cairo.Context`): The canvas on which to render the drawings
        """
        ww, wh = widget.get_allocated_width(), widget.get_allocated_height()

        if self.zoom_selecting and self.zoom_points:
            xmin, xmax = sorted(p[0] * ww for p in self.zoom_points)
            ymin, ymax = sorted(p[1] * wh for p in self.zoom_points)

            rect = Gdk.Rectangle()
            rect.x = xmin
            rect.width = xmax - xmin
            rect.y = ymin
            rect.height = ymax - ymin

            cairo_context.set_line_width(3)
            cairo_context.set_line_cap(cairo.LINE_CAP_SQUARE)
            Gdk.cairo_rectangle(cairo_context, rect)
            cairo_context.set_source_rgba(.1, .1, 1, .4)
            cairo_context.stroke()

            Gdk.cairo_rectangle(cairo_context, rect)
            cairo_context.set_source_rgba(.5, .5, 1, .2)
            cairo_context.fill()


    def nop(*args, **kwargs):
        """ Do nothing
        """
        pass



class FileWatcher(object):
    """ A class that wraps watchdog objects, to trigger callbacks when a file changes.
    """
    #: A :class:`~watchdog.observers.Observer` to watch when the file changes
    observer = None

    #: A :class:`~watchdog.events.FileSystemEventHandler` to get notified when the file changes
    monitor = None

    # `int` that is a GLib timeout id to delay the callback
    timeout = 0

    #: The :class:`~pathlib.Path` to the file being watched
    path = None

    #: Callback to be called on file changes, usually connected to :meth:`~pympress.ui.UI.reload_document`
    callback = lambda: None

    def __init__(self):
        try:
            from watchdog.observers import Observer
            from watchdog.events import FileSystemEventHandler

            self.observer = Observer()
            self.monitor = FileSystemEventHandler()

            self.observer.start()

        except ImportError:
            logger.error(_('Missing dependency: python "{}" package').format('watchdog'))
            logger.info(_('Monitoring of changes to reload files automatically is not available'))


    def __del__(self):
        """ On finalize, cancel the watchdog observer thread.
        """
        self.stop_watching()
        if self.observer.is_alive():
            self.observer.stop()

        self.observer = None


    def watch_file(self, uri, callback, *args, **kwargs):
        """ Watches a new file with a new callback. Removes any precedent watched files.

        If the optional watchdog dependency is missing, does nothing.

        Args:
            uri (`str`): URI of the file to watch
            callback (`function`): callback to call with all the further arguments when the file changes
        """
        if self.observer is None:
            return

        self.stop_watching()

        scheme, path = uri.split('://', 1)
        self.path = pathlib.Path(url2pathname(path))
        if scheme != 'file':
            logger.error('Impossible to watch files with {} schemes'.format(scheme), exc_info = True)
            return

        self.callback = lambda: callback(*args, **kwargs)
        self.monitor.on_modified = self._enqueue
        try:
            self.observer.schedule(self.monitor, str(self.path.parent), recursive=False)
        except OSError:
            logger.error('Impossible to open dir at {}'.format(str(self.path.parent)), exc_info = True)


    def stop_watching(self):
        """ Remove all files that are being watched.
        """
        self.observer.unschedule_all()


    def _enqueue(self, event):
        """ Call callback with delay, to avoid repeated calls in short periods of time.

        Args:
            event (:class:`~watchdog.events.FileSystemEvent`): the event that caused the callback to be triggered
        """
        if event.src_path != str(self.path):
            return
        if self.timeout:
            GLib.Source.remove(self.timeout)
        self.timeout = GLib.timeout_add(200, self._call)


    def _call(self):
        """ Call the callback.
        """
        if self.timeout:
            self.timeout = 0
        self.callback()



# -*- coding: utf-8 -*-
#
#       editable_label.py
#
#       Copyright 2017 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.editable_label` -- A label that can be swapped out for an editable entry
---------------------------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib, Gio


class EditableLabel(object):
    """ A label that can switch between simply displaying a value, and allowing user input to edit this value.
    """
    #: :class:`~Gtk.EventBox` around the label, used to sense clicks
    event_box = None

    #: `bool` tracking whether we are currently editing the label.
    editing = False

    def on_label_event(self, widget_or_action, event=None):
        """ Manage events on the current slide label/entry.

        This function triggers replacing the label with an entry when clicked or otherwise toggled.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget in which the event occurred
            event (:class:`~Gtk.Event` or None):  the event that occurred, None if tf we called from a menu item

        Returns:
            `bool`: whether the event was consumed
        """
        hint = None
        if issubclass(type(widget_or_action), Gio.Action):
            hint = widget_or_action.get_name()

        elif event is not None and event.type == Gdk.EventType.BUTTON_PRESS:
            # If we clicked on the Event Box then don't toggle, just enable.
            if widget_or_action is not self.event_box or self.editing:
                return False

        else:
            return False

        # Perform the state toggle
        if not self.editing:
            self.swap_label_for_entry(hint)
        else:
            self.validate()

        return True


    def validate(self):
        """ Validate the input to the label. Needs to be reimplemented by children classes.
        """
        raise NotImplementedError


    def cancel(self):
        """ Cancel editing the label. Needs to be reimplemented by children classes.
        """
        pass


    def on_keypress(self, widget, event):
        """ Manage key presses for the editable label. Needs to be reimplemented by children classes.

        If we are editing the label, intercept some key presses (to validate or cancel editing or other specific
        behaviour), otherwise pass the key presses on to the button for normal behaviour.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.
            name (`str`): the name of the key stroke
            command (`str`): the name of the command in case this function is called by on_key_input

        Returns:
            `bool`: whether the event was consumed
        """
        raise NotImplementedError


    def try_cancel(self):
        """ Cancel editing the label, if it is being edited.

        Returns:
            `bool`: `True` if editing got cancelled, `False` if the label was not being edited.
        """
        if not self.editing:
            return False

        self.cancel()
        self.restore_label()
        return True


    def try_validate(self):
        """ Validate the page choice, if the page label is being edited.

        Returns:
            `bool`: `True` if editing got validated, `False` if the label was not being edited.
        """
        if not self.editing:
            return False

        self.validate()
        self.restore_label()
        return True


    def swap_label_for_entry(self):
        """ Perform the actual work of starting the editing.
        """
        raise NotImplementedError

    def restore_label(self):
        """ Make sure that the editable label is not in entry mode.

        If it is an entry, then replace it with the label.
        """
        raise NotImplementedError


    def start_editing(self):
        """ Start the editing of the label if it is disabled.
        """
        if not self.editing:
            self.swap_label_for_entry()


    def stop_editing(self):
        """ Disable the editing of the label if it was enabled.
        """
        if self.editing:
            self.restore_label()


class PageNumber(EditableLabel):
    """ A label that displays "current page / max page", that can be edited to select a page to which to go.

    Args:
        builder (:class:`~pympress.builder.Builder`): A builder from which to load widgets
    """
    #: Slide counter :class:`~Gtk.Label` for the current slide.
    label_cur = None
    #: Slide counter :class:`~Gtk.Label` for the last slide.
    label_last = None
    #: :class:`~Gtk.EventBox` associated with the slide counter label in the Presenter window.
    eb_cur = None
    #: :class:`~Gtk.HBox` containing the slide counter label in the Presenter window.
    hb_cur = None
    #: :class:`~Gtk.SpinButton` used to switch to another slide by typing its number.
    spin_cur = None
    #: :class:`~Gtk.Entry` used to switch to another slide by typing its label.
    edit_label = None
    #: :class:`~Gtk.Label` separating `~spin_cur` and `~edit_label`
    label_sep = None

    #: `int` holding the maximum page number in the document
    max_page_number = 1
    #: `bool` holding whether we display or ignore page labels
    page_labels = True
    #: `bool` whether to scroll with the pages (True) or with the page numbers (False)
    invert_scroll = True

    #: callback, to be connected to :func:`~pympress.document.Document.goto`
    goto_page = lambda *args: None
    #: callback, to be connected to :func:`~pympress.document.Document.lookup_label`
    find_label = lambda *args: None
    #: callback, to be connected to :func:`~pympress.document.Document.label_after`
    label_before = lambda *args: None
    #: callback, to be connected to :func:`~pympress.document.Document.label_before`
    label_after = lambda *args: None
    #: callback, to be connected to :func:`~pympress.ui.UI.do_page_change`
    page_change = lambda *args: None

    def __init__(self, builder, page_num_scroll):
        super(PageNumber, self).__init__()

        # The spinner's scroll is with page numbers, invert to scroll with pages
        self.invert_scroll = not page_num_scroll

        builder.load_widgets(self)
        builder.setup_actions({
            'goto-page':          dict(activate=self.on_label_event),
            'jumpto-label':       dict(activate=self.on_label_event),
            'cancel-page-number': dict(activate=self.cancel),
        })

        self.goto_page       = builder.get_callback_handler('goto_page')
        self.page_change     = builder.get_callback_handler('do_page_change')
        self.setup_doc_callbacks(builder.doc)

        # Initially (from XML) both the spinner and the current page label are visible.
        self.hb_cur.remove(self.spin_cur)
        self.hb_cur.remove(self.edit_label)
        self.hb_cur.remove(self.label_sep)

        self.event_box = self.eb_cur


    def setup_doc_callbacks(self, doc):
        """ Callbacks that need to be setup again at every new document

        Args:
            doc (:class:`~pympress.document.Document`): The new document that got loaded
        """
        self.find_label   = doc.lookup_label
        self.label_after  = doc.label_after
        self.label_before = doc.label_before


    def set_last(self, num_pages):
        """ Set the max number of pages, both on display and as the range of values for the spinner.

        Args:
            num_pages (`int`): The maximum page number
        """
        self.max_page_number = num_pages
        self.label_last.set_text(('/{})' if self.page_labels else '/{}').format(num_pages))
        self.spin_cur.set_range(1, num_pages)
        self.spin_cur.set_max_length(len(str(num_pages)) + 1)


    def enable_labels(self, enable):
        """ Allow one to use or ignore labels.

        Args:
            enable (`bool`): Whether to enable labels
        """
        self.page_labels = enable
        self.label_last.set_text(('/{})' if enable else '/{}').format(self.max_page_number))


    def changed_page_label(self, *args):
        """ Get the page number from the spinner and go to that page.
        """
        if not self.page_labels or not self.edit_label.is_focus() or not self.edit_label.get_text():
            return

        page_nb = self.find_label(self.edit_label.get_text(), prefix_unique = True)
        if not page_nb:
            return

        # use the spinner's mechanism
        self.spin_cur.set_value(page_nb + 1)


    def validate(self):
        """ Get the page number from the spinner and go to that page.
        """
        page_nb = None
        if self.page_labels and self.edit_label.is_focus():
            page_nb = self.find_label(self.edit_label.get_text(), prefix_unique = False)

        if page_nb is None:
            page_nb = int(self.spin_cur.get_value() - 1)

        if page_nb is not None:
            self.restore_label()
            self.goto_page(page_nb)
        else:
            self.cancel()


    def cancel(self, gaction=None, param=None):
        """ Make the UI re-display the pages from before editing the current page.
        """
        self.restore_label()
        GLib.idle_add(self.page_change, unpause=False)


    def on_keypress(self, widget, event):
        """ Implement directions (left/right/home/end) keystrokes.

        Otherwise pass on to :func:`~Gtk.SpinButton.do_key_press_event()`.
        """
        if self.page_labels and self.edit_label.is_focus():
            return Gtk.Entry.do_key_press_event(self.edit_label, event)
        else:
            return Gtk.SpinButton.do_key_press_event(self.spin_cur, event)


    def on_scroll(self, widget, event):
        """ Scroll event. Pass it on to the spin button if we're currently editing the page number.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.

        Returns:
            `bool`: whether the event was consumed
        """
        if not self.editing:
            return False
        else:
            # flip scroll direction to get scroll down advancing slides
            if self.invert_scroll and event.direction == Gdk.ScrollDirection.DOWN:
                event.direction = Gdk.ScrollDirection.UP
            elif self.invert_scroll and event.direction == Gdk.ScrollDirection.UP:
                event.direction = Gdk.ScrollDirection.DOWN

            # Manually get destination slide if we're editing labels
            if self.edit_label.is_focus():
                cur_page = int(self.spin_cur.get_value()) - 1
                if event.direction == Gdk.ScrollDirection.DOWN:
                    self.spin_cur.set_value(1 + self.label_before(cur_page))
                elif event.direction == Gdk.ScrollDirection.UP:
                    self.spin_cur.set_value(1 + self.label_after(cur_page))

            # Otherwise let the spinner do its job
            else:
                return Gtk.SpinButton.do_scroll_event(self.spin_cur, event)


    def swap_label_for_entry(self, hint=None):
        """ Perform the actual work of starting the editing.
        """
        Gio.Application.get_default().activate_action('cancel-talk-time')

        label, sep, cur = self.label_cur.get_text().rpartition('(')

        # Replace label with entry
        self.spin_cur.show()
        self.hb_cur.pack_start(self.spin_cur, True, True, 0)
        self.hb_cur.reorder_child(self.spin_cur, 1)

        if self.page_labels:
            self.hb_cur.pack_start(self.edit_label, True, True, 0)
            self.hb_cur.reorder_child(self.edit_label, 0)
            self.edit_label.set_text(label.strip())

            self.hb_cur.pack_start(self.label_sep, True, True, 0)
            self.hb_cur.reorder_child(self.label_sep, 1)
            self.label_sep.set_text(' (')

            self.hb_cur.set_homogeneous(False)

        self.hb_cur.remove(self.label_cur)

        try:
            cur_nb = int(cur.strip())
        except ValueError:
            cur_nb = -1
        self.spin_cur.set_value(cur_nb)

        if self.page_labels and (hint == 'jumpto-label' or hint == 'nav_jump'):
            self.edit_label.grab_focus()
            self.edit_label.select_region(0, -1)
        else:
            self.spin_cur.grab_focus()
            self.spin_cur.select_region(0, -1)

        self.editing = True


    def restore_label(self):
        """ Make sure that the current page number is displayed in a label and not in an entry.

        If it is an entry, then replace it with the label.
        """
        if self.spin_cur in self.hb_cur:
            if self.page_labels:
                self.hb_cur.set_homogeneous(True)
                self.hb_cur.remove(self.edit_label)
                self.hb_cur.remove(self.label_sep)
            self.hb_cur.remove(self.spin_cur)
            self.hb_cur.pack_start(self.label_cur, True, True, 0)
            self.hb_cur.reorder_child(self.label_cur, 0)

        self.editing = False


    def update_page_numbers(self, cur_nb, label):
        """ Update the displayed page numbers.

        Args:
            cur_nb (`int`): The current page number, in documentation numbering (range [0..max - 1])
            label (`str`): The current page label
        """
        cur = str(cur_nb + 1)

        if self.page_labels:
            self.label_cur.set_text('{} ({}'.format(str(label), cur))
        else:
            self.label_cur.set_text(cur)

        if self.editing:
            if not self.edit_label.is_focus():
                self.edit_label.set_text(str(label))

            self.spin_cur.set_value(cur_nb + 1)


class EstimatedTalkTime(EditableLabel):
    """ A label that displays the time elapsed since the start of the talk, that can be edited to select talk duration.

    The duration of the talk will cause the label to blink and change colour as the elapsed time gets closer to the
    targeted talk duration.

    Args:
        builder (builder.Builder): The builder from which to load widgets.
    """
    #: Elapsed time :class:`~Gtk.Label`.
    label_time = None
    #: Estimated talk time :class:`~Gtk.Label` for the talk.
    label_ett = None
    #: :class:`~Gtk.EventBox` associated with the estimated talk time.
    eb_ett = None
    #: Estimated talk time, `int` in seconds.
    est_time = 0

    #: :class:`~Gtk.Entry` used to set the estimated talk time.
    entry_ett = None

    #: callback, to be connected to :func:`~pympress.editable_label.PageNumber.stop_editing`
    stop_editing_page_number = lambda *args: None


    def __init__(self, builder):
        super(EstimatedTalkTime, self).__init__()

        self.entry_ett = Gtk.Entry()

        builder.load_widgets(self)
        builder.setup_actions({
            'edit-talk-time':   dict(activate=self.on_label_event),
            'set-talk-time':    dict(activate=self.set_time, parameter_type=int),
            'cancel-talk-time': dict(activate=self.restore_label),
        })

        self.event_box = self.eb_ett


    def validate(self):
        """ Update estimated talk time from the input.
        """
        text = self.entry_ett.get_text()

        t = ["0" + n.strip() for n in text.split(':')]
        try:
            m = int(t[0])
            s = int(t[1])
        except ValueError:
            logger.error(_("Invalid time (mm or mm:ss expected), got \"{}\"").format(text))
            return
        except IndexError:
            s = 0

        self.set_time(None, GLib.Variant.new_int64(m * 60 + s))


    def set_time(self, gaction, param):
        """ Set the talk time.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): The time in seconds, as an int64 variant
        """
        time = param.get_int64()
        self.est_time = time
        self.label_ett.set_text("{:02}:{:02}".format(*divmod(self.est_time, 60)))


    def on_keypress(self, widget, event):
        """ Pass on keystrokes to :func:`~Gtk.Entry.do_key_press_event()`.
        """
        return Gtk.Entry.do_key_press_event(self.entry_ett, event)


    def swap_label_for_entry(self, *args):
        """ Perform the actual work of starting the editing.
        """
        Gio.Application.get_default().activate_action('cancel-page-number')

        # Set entry text
        self.entry_ett.set_text("{:02}:{:02}".format(*divmod(self.est_time, 60)))
        self.entry_ett.select_region(0, -1)

        # Replace label with entry
        self.eb_ett.remove(self.label_ett)
        self.eb_ett.add(self.entry_ett)
        self.entry_ett.show()
        self.entry_ett.grab_focus()
        self.editing = True


    def restore_label(self, gaction=None, param=None):
        """ Make sure that the current page number is displayed in a label and not in an entry.

        If it is an entry, then replace it with the label.
        """
        child = self.eb_ett.get_child()
        if child is not self.label_ett:
            self.eb_ett.remove(child)
            self.eb_ett.add(self.label_ett)

        self.editing = False



# -*- coding: utf-8 -*-
#
#       pympress
#
#       Copyright 2009, 2010 Thomas Jost <thomas.jost@gmail.com>
#       Copyright 2015 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.__main__` -- The entry point of pympress
-------------------------------------------------------
"""

import logging
import os
import sys
import locale
import gettext

from pympress import util


# Setup logging, and catch all uncaught exceptions in the log file.
# Load pympress.util early (OS and path-specific things) to load and setup gettext translation asap.
logger = logging.getLogger(__name__)
logging.basicConfig(filename=util.get_log_path(), level=logging.DEBUG)


def uncaught_handler(*exc_info):
    """ Exception handler, to log uncaught exceptions to our log file.
    """
    logger.critical('Uncaught exception:\n{}'.format(logging.Formatter().formatException(exc_info)))
    sys.__excepthook__(*exc_info)


sys.excepthook = uncaught_handler


if util.IS_WINDOWS:
    if os.getenv('LANG') is None:
        lang, enc = locale.getdefaultlocale()
        os.environ['LANG'] = lang

try:
    loaded_locale = locale.setlocale(locale.LC_ALL, '')
except locale.Error as err:
    logger.exception('Failed loading locale: {}'.format(err))
    print('Failed loading locale: {}'.format(err), file=sys.stderr)

gettext.install('pympress', util.get_locale_dir())


try:
    # python <3.6 does not have this
    ModuleNotFoundError
except NameError:
    ModuleNotFoundError = ImportError


# Load python bindings for gobject introspections, aka pygobject, aka gi, and pycairo.
# These are dependencies that are not specified in the setup.py, so we need to start here.
# They are not specified because:
# - installing those via pip requires compiling (always for pygobject, if no compatible wheels exist for cairo),
# - compiling requires a compiling toolchain, development packages of the libraries, etc.,
# - all of this makes more sense to be handled by the OS package manager,
# - it is hard to make pretty error messages pointing this out at `pip install` time,
#   as they would have to be printed when the dependency resolution happens.
# See https://github.com/Cimbali/pympress/issues/100
try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk, Gdk, GLib, Gio
    import cairo
except ModuleNotFoundError:
    logger.critical('Gobject Introspections and/or pycairo module is missing', exc_info = True)
    print('\n' + _('ERROR: Gobject Introspections and/or pycairo module is missing, ' +
                   'make sure Gtk, pygobject and pycairo are installed on your system.') + '\n')
    print(_('Try your operating system’s package manager, or try running: pip install pygobject pycairo'))
    print(_('pip will then download and compile pygobject and pycairo, ' +
            'for which you need the Gtk and cairo headers (or development packages).') + '\n')
    print(_('For instructions, refer to https://github.com/Cimbali/pympress/blob/master/README.md#dependencies'))
    print(_('If using a virtualenv or anaconda, you can also try allowing system site packages.'))
    print()
    exit(1)



# Finally the real deal: load pympress modules, handle command line args, and start up
from pympress import app


def main(argv = sys.argv[:]):
    """ Entry point of pympress. Parse command line arguments, instantiate the UI, and start the main loop.
    """
    app.Pympress().run(argv)


if __name__ == "__main__":
    main()

##
# Local Variables:
# mode: python
# indent-tabs-mode: nil
# py-indent-offset: 4
# fill-column: 80
# end:



# -*- coding: utf-8 -*-
#
#       pympress
#
#       Copyright 2009, 2010 Thomas Jost <thomas.jost@gmail.com>
#       Copyright 2015 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.app` -- The Gtk.Application managing the lifetime and CLI
------------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import signal
import platform

import gi
import cairo
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib, Gio

from pympress import util, config, document, ui, builder


class Pympress(Gtk.Application):
    """ Class representing the single pympress Gtk application.
    """
    #: The :class:`~pympress.ui.UI` object that is the interface of pympress
    gui = None
    #: The :class:`~pympress.config.Config` object that holds pympress conferences
    config = None
    #: `list` of actions to be passsed to the GUI that were queued before GUI was created
    action_startup_queue = []
    #: `bool` to automatically upgrade log level (DEBUG / INFO at init, then ERROR), False if user set log level
    auto_log_level = True

    options = {
        # long_name:  (short_name (int), flags (GLib.OptionFlags), arg (GLib.OptionArg)
        'talk-time':  (ord('t'), GLib.OptionFlags.NONE, GLib.OptionArg.STRING),
        'notes':      (ord('N'), GLib.OptionFlags.NONE, GLib.OptionArg.STRING),
        'log':        (0,        GLib.OptionFlags.NONE, GLib.OptionArg.STRING),
        'version':    (ord('v'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE),
        'pause':      (ord('P'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE),
        'reset':      (ord('r'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE),
        'next':       (ord('n'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE),
        'prev':       (ord('p'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE),
        'first':      (ord('f'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE),
        'last':       (ord('l'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE),
        'blank':      (ord('b'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE),
        'quit':       (ord('q'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE),
    }

    option_descriptions = {
        #  long_name: (description, arg_description)
        'talk-time': (_('The estimated (intended) talk time in minutes') + ' ' +
                      _('(and optionally seconds)'), 'mm[:ss]'),
        'notes':     (_('Set the position of notes on the pdf page') + ' ' +
                      _('(none, left, right, top, bottom, after, odd, or prefix).') + ' ' +
                      _('Overrides the detection from the file.'), '<position>'),
        'log':       (_('Set level of verbosity in log file:') + ' ' +
                      _('{}, {}, {}, {}, or {}').format('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'), '<level>'),
        'version':   (_('Print version and exit'), None),
        'pause':     (_('Toggle pause of talk timer'), None),
        'reset':     (_('Reset talk timer'), None),
        'next':      (_('Next slide'), None),
        'prev':      (_('Previous slide'), None),
        'first':     (_('First slide'), None),
        'last':      (_('Last slide'), None),
        'blank':     (_('Blank/unblank content screen'), None),
        'quit':      (_('Close opened pympress instance'), None),
    }

    version_string = ' '.join([
        'Pympress:', util.get_pympress_meta()['version'],
        '; Python:', platform.python_version(),
        '; OS:', platform.system(), platform.release(), platform.version(),
        '; Gtk {}.{}.{}'.format(Gtk.get_major_version(), Gtk.get_minor_version(), Gtk.get_micro_version()),
        '; GLib {}.{}.{}'.format(GLib.MAJOR_VERSION, GLib.MINOR_VERSION, GLib.MICRO_VERSION),
        '; Poppler', document.Poppler.get_version(), document.Poppler.get_backend().value_nick,
        '; Cairo', cairo.cairo_version_string(), ', pycairo', cairo.version,
    ])


    def __init__(self):
        GLib.set_application_name('pympress')
        # GLib.set_prgname('pympress')  # Let prgname be auto-determined from sys.argv[0]
        Gtk.Application.__init__(self, application_id='io.github.pympress',
                                 flags=Gio.ApplicationFlags.HANDLES_OPEN | Gio.ApplicationFlags.CAN_OVERRIDE_APP_ID)

        self.register(None)

        if not self.get_is_remote():
            builder.Builder.setup_actions({
                'log-level': dict(activate=self.set_log_level, state=logger.getEffectiveLevel(), parameter_type=int),
            }, action_map=self)

        # Connect proper exit function to interrupt
        signal.signal(signal.SIGINT, self.quit)

        for opt in self.options:
            self.add_main_option(opt, *self.options[opt], *self.option_descriptions.get(opt, ['', None]))


    def quit(self, *args):
        """ Quit and ignore other arguments e.g. sent by signals.
        """
        if self.gui is not None and self.gui.unsaved_changes():
            return

        Gtk.Application.quit(self)
        return False


    def do_startup(self):
        """ Common start-up tasks for primary and remote instances.

        NB. super(self) causes segfaults, Gtk.Application needs to be used as base.
        """
        self.config = config.Config()

        # prefere X11 on posix systems because Wayland still has some shortcomings for us,
        # specifically libVLC and the ability to disable screensavers
        if util.IS_POSIX:
            Gdk.set_allowed_backends('x11,*')

        logger.info(self.version_string)
        Gtk.Application.do_startup(self)


    def do_activate(self, timestamp=GLib.get_current_time()):
        """ Activate: show UI windows.

        Build them if they do not exist, otherwise bring to front.
        """
        if self.gui is None:
            if self.auto_log_level:
                self.activate_action('log-level', logging.INFO)
                self.action_startup_queue.append(('log-level', logging.ERROR))

            # Build the UI and windows
            self.gui = ui.UI(self, self.config)

            while self.action_startup_queue:
                self.activate_action(*self.action_startup_queue.pop(0))

        Gtk.Application.do_activate(self)
        self.gui.p_win.present_with_time(timestamp)


    def set_action_enabled(self, name, value):
        """ Parse an action name and set its enabled state to True or False.

        Args:
            name (`str`): the name of the stateful action
            value (`bool`): wheether the action should be enabled or disabled
        """
        self.lookup_action(name).set_enabled(value)


    def set_action_state(self, name, value):
        """ Parse an action name and set its state wrapped in a :class:`~GLib.Variant`.

        Args:
            name (`str`): the name of the stateful action
            value (`str`, `int`, `bool` or `float`): the value to set.
        """
        self.lookup_action(name).change_state(GLib.Variant(builder.Builder._glib_type_strings[type(value)], value))


    def get_action_state(self, name):
        """ Parse an action name and return its unwrapped state from the :class:`~GLib.Variant`.

        Args:
            name (`str`): the name of the stateful action

        Returns:
            `str`, `int`, `bool` or `float`: the value contained in the action
        """
        state = self.lookup_action(name).get_state()
        return builder.Builder._glib_type_getters[state.get_type_string()](state)


    def activate_action(self, name, parameter=None):
        """ Parse an action name and activate it, with parameter wrapped in a :class:`~GLib.Variant` if it is not None.

        Args:
            name (`str`): the name of the stateful action
            parameter: an object or None to pass as a parameter to the action, wrapped in a GLib.Variant
        """
        if not self.get_is_remote() and self.gui is None and name not in ['log-level']:
            self.action_startup_queue.append((name, parameter))
            return

        if parameter is not None:
            parameter = GLib.Variant(builder.Builder._glib_type_strings[type(parameter)], parameter)

        Gio.ActionGroup.activate_action(self, name, parameter)


    def do_open(self, files, n_files, hint):
        """ Handle opening files. In practice we only open once, the last one.

        Args:
            files (`list` of :class:`~Gio.File`): representing an array of files to open
            n_files (`int`): the number of files passed.
            hint (`str`): a hint, such as view, edit, etc. Should always be the empty string.
        """
        if not n_files:
            return

        self.do_activate(timestamp=GLib.get_current_time())
        self.gui.swap_document(files[-1].get_uri())


    def do_shutdown(self):
        """ Perform various cleanups and save preferences.
        """
        if self.gui is not None:
            self.gui.cleanup()

        self.config.save_config()
        Gtk.Application.do_shutdown(self)


    def set_log_level(self, action, param):
        """ Action that sets the logging level (on the root logger of the active instance)

        Args:
            action (:class:`~Gio.Action`): The action activatd
            param (:class:~`GLib.Variant`): The desired level as an int wrapped in a GLib.Variant
        """
        logging.getLogger(None).setLevel(param.get_int64())
        action.change_state(param)


    def do_handle_local_options(self, opts_variant_dict):
        """ Parse command line options, returned as a VariantDict

        Returns:
            `tuple`: estimated talk time, log level, notes positions.
        """
        # convert GVariantDict -> GVariant -> dict
        opts = opts_variant_dict.end().unpack()

        simple_actions = {
            'pause': 'pause-timer',
            'reset': 'reset-timer',
            'next': 'next-page',
            'prev': 'prev-page',
            'blank': 'blank-screen',
            'quit': 'quit',
            'first': 'first-page',
            'last': 'last-page',
        }

        for opt, arg in opts.items():
            if opt == "version":
                print(self.version_string)
                return 0

            elif opt == "log":
                numeric_level = getattr(logging, arg.upper(), None)
                if isinstance(numeric_level, int):
                    self.auto_log_level = False
                    self.activate_action('log-level', numeric_level)
                else:
                    print(_("Invalid log level \"{}\", try one of {}").format(
                        arg, "DEBUG, INFO, WARNING, ERROR, CRITICAL"
                    ))

            elif opt == "notes":
                arg = arg.lower()[:1]
                if arg == 'n': self.activate_action('notes-pos', 'none')
                if arg == 'l': self.activate_action('notes-pos', 'left')
                if arg == 'r': self.activate_action('notes-pos', 'right')
                if arg == 't': self.activate_action('notes-pos', 'top')
                if arg == 'b': self.activate_action('notes-pos', 'bottom')
                if arg == 'a': self.activate_action('notes-pos', 'after')
                if arg == 'o': self.activate_action('notes-pos', 'odd')
                if arg == 'p': self.activate_action('notes-pos', 'map')  # prefix -> map

            elif opt == "talk-time":
                t = ["0" + n.strip() for n in arg.split(':')]
                try:
                    m = int(t[0])
                    s = int(t[1])
                except ValueError:
                    print(_("Invalid time (mm or mm:ss expected), got \"{}\"").format(arg))
                    return 2
                except IndexError:
                    s = 0
                self.activate_action('set-talk-time', m * 60 + s)

            elif opt in simple_actions:
                self.activate_action(simple_actions[opt])

        return -1



# -*- coding: utf-8 -*-
#
#       talk_time.py
#
#       Copyright 2017 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.talk_time` -- Manages the clock of elapsed talk time
-------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import time

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib


class TimeLabelColorer(object):
    """ Manage the colors of a label with a set of colors between which to fade, based on how much time remains.

    Times are given in seconds (<0 has run out of time). In between timestamps the color will interpolated linearly,
    outside of the intervals the closest color will be used.

    Args:
        label_time (:class:`Gtk.Label`): the label where the talk time is displayed
    """

    #: The :class:`Gtk.Label` whose colors need updating
    label_time = None

    #: :class:`~Gdk.RGBA` The default color of the info labels
    label_color_default = None

    #: :class:`~Gtk.CssProvider` affecting the style context of the labels
    color_override = None

    #: `list` of tuples (`int`, :class:`~Gdk.RGBA`), which are the desired colors at the corresponding timestamps.
    #: Sorted on the timestamps.
    color_map = []

    def __init__(self, label_time):
        self.label_time = label_time

        style_context = self.label_time.get_style_context()
        self.color_override = Gtk.CssProvider()
        style_context.add_provider(self.color_override, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION + 1)

        self.label_color_default = self.load_color_from_css(style_context)
        label_color_ett_reached = self.load_color_from_css(style_context, "ett-reached")
        label_color_ett_info = self.load_color_from_css(style_context, "ett-info")
        label_color_ett_warn = self.load_color_from_css(style_context, "ett-warn")

        self.color_map = [
            ( 300, self.label_color_default),
            (   0, label_color_ett_reached),
            (-150, label_color_ett_info),
            (-300, label_color_ett_warn)
        ]


    def load_color_from_css(self, style_context, class_name = None):
        """ Add class class_name to the time label and return its color.

        Args:
            label_time (:class:`Gtk.Label`): the label where the talk time is displayed
            style_context (:class:`~Gtk.StyleContext`): the CSS context managing the color of the label
            class_name (`str` or `None`): The name of the class, if any

        Returns:
            :class:`~Gdk.RGBA`: The color of the label with class "class_name"
        """
        if class_name:
            style_context.add_class(class_name)

        self.label_time.show()
        color = style_context.get_color(Gtk.StateType.NORMAL)

        if class_name:
            style_context.remove_class(class_name)

        return color


    def default_color(self):
        """ Forces to reset the default colors on the label.
        """
        self.color_override.load_from_data(''.encode('ascii'))


    def update_time_color(self, remaining):
        """ Update the color of the time label based on how much time is remaining.

        Args:
            remaining (`int`): Remaining time until estimated talk time is reached, in seconds.
        """
        if (remaining <= 0 and remaining > -5) or (remaining <= -300 and remaining > -310):
            self.label_time.get_style_context().add_class("time-warn")
        else:
            self.label_time.get_style_context().remove_class("time-warn")

        prev_time, prev_color = None, None

        for timestamp, color in self.color_map:
            if remaining >= timestamp:
                break
            prev_time, prev_color = (timestamp, color)

        else:
            # if remaining < all timestamps, use only last color
            prev_color = None

        if prev_color:
            position = (remaining - prev_time) / (timestamp - prev_time)
            color_spec = '* {{color: mix({}, {}, {})}}'.format(prev_color.to_string(), color.to_string(), position)
        else:
            color_spec = '* {{color: {}}}'.format(color.to_string())

        self.color_override.load_from_data(color_spec.encode('ascii'))


class TimeCounter(object):
    """ A double counter, that displays the time elapsed in the talk and a clock.

    Args:
        builder (builder.Builder): The builder from which to load widgets.
        ett (`int`): the estimated time for the talk, in seconds.
        timing_tracker: (:class:`~pympress.extras.TimingReport`): to inform when the slides change
        autoplay: (:class:`~pympress.dialog.AutoPlay`): to adjust the timer display if we’re auto-playing/looping slides
    """
    #: Elapsed time :class:`~Gtk.Label`
    label_time = None
    #: Clock :class:`~Gtk.Label`
    label_clock = None

    #: Time at which the counter was started, `int` in seconds as returned by :func:`~time.time()`
    restart_time = 0
    #: Time elapsed since the beginning of the presentation, `int` in seconds
    elapsed_time = 0
    #: Timer paused status, `bool`
    paused = True

    #: :class:`~TimeLabelColorer` that handles setting the colors of :attr:`label_time`
    label_colorer = None

    #: :class:`~pympress.editable_label.EstimatedTalkTime` that handles changing the ett
    ett = None

    #: The pause-timer :class:`~Gio.Action`
    pause_action = None

    #: The :class:`~pympress.extras.TimingReport`, needs to know when the slides change
    timing_tracker = None
    #: The :class:`~pympress.dialog.AutoPlay`, to adjust the timer display if we’re auto-playing/looping slides
    autoplay = None

    def __init__(self, builder, ett, timing_tracker, autoplay):
        super(TimeCounter, self).__init__()

        self.label_colorer = TimeLabelColorer(builder.get_object('label_time'))
        self.ett = ett
        self.timing_tracker = timing_tracker
        self.autoplay = autoplay

        builder.load_widgets(self)

        builder.setup_actions({
            'pause-timer':    dict(activate=self.switch_pause, state=self.paused),
            'reset-timer':    dict(activate=self.reset_timer),
        })
        self.pause_action = builder.get_application().lookup_action('pause-timer')

        # Setup timer for clocks
        GLib.timeout_add(250, self.update_time)


    def switch_pause(self, gaction, param=None):
        """ Switch the timer between paused mode and running (normal) mode.

        Returns:
            `bool`: whether the clock's pause was toggled.
        """
        if self.paused:
            self.unpause()
        else:
            self.pause()
        return None


    def pause(self):
        """ Pause the timer if it is not paused, otherwise do nothing.

        Returns:
            `bool`: whether the clock's pause was toggled.
        """
        if self.paused:
            return False

        self.paused = True
        self.pause_action.change_state(GLib.Variant.new_boolean(self.paused))

        self.elapsed_time += time.time() - self.restart_time
        self.timing_tracker.end_time = self.elapsed_time

        if self.autoplay.is_looping():
            self.autoplay.pause()

        self.update_time()
        return True


    def unpause(self):
        """ Unpause the timer if it is paused, otherwise do nothing.

        Returns:
            `bool`: whether the clock's pause was toggled.
        """
        if not self.paused:
            return False

        self.restart_time = time.time()

        self.paused = False
        self.pause_action.change_state(GLib.Variant.new_boolean(self.paused))

        if self.autoplay.is_looping():
            self.autoplay.unpause()

        self.update_time()
        return True


    def reset_timer(self, *args):
        """ Reset the timer.
        """
        self.timing_tracker.reset(self.current_time())

        self.restart_time = time.time()
        self.elapsed_time = 0
        if self.autoplay.is_looping():
            self.autoplay.start_looping()
        self.update_time()


    def current_time(self):
        """ Returns the time elapsed in the presentation.

        Returns:
            `int`: the time since the presentation started in seconds.
        """
        # Time elapsed since the beginning of the presentation
        if self.paused:
            return self.elapsed_time
        else:
            return self.elapsed_time + (time.time() - self.restart_time)


    def update_time(self):
        """ Update the timer and clock labels.

        Returns:
            `bool`: `True` (to prevent the timer from stopping)
        """
        # Current time
        clock = time.strftime('%X')  # '%H:%M:%S'
        elapsed = self.current_time()

        # Time elapsed since the beginning of the presentation
        if self.autoplay.is_looping():
            first, stop, loop, delay = self.autoplay.get_page_range()
            display_time = '{} {}-{} / {:.1f}s'.format(_('Loop') if loop else _('Auto'), first + 1, stop, delay / 1000)
        else:
            display_time = '{:02}:{:02}'.format(*divmod(int(elapsed), 60))

        if self.paused:
            display_time += ' ' + _('(paused)')

        self.label_time.set_text(display_time)
        self.label_clock.set_text(clock)
        if not self.paused:
            self.timing_tracker.end_time = elapsed

        if self.ett.est_time:
            self.label_colorer.update_time_color(self.ett.est_time - elapsed)
        else:
            self.label_colorer.default_color()

        return True



# -*- coding: utf-8 -*-
#
#       document.py
#
#       Copyright 2009, 2010 Thomas Jost <thomas.jost@gmail.com>
#       Copyright 2015 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.document` -- document handling
---------------------------------------------

This module contains several classes that are used for managing documents (only
PDF documents are supported at the moment, but other formats may be added in the
future).

An important point is that this module is *completely* independent from the GUI:
there should not be any GUI-related code here, except for page rendering (and
only rendering itself: the preparation of the target surface must be done
elsewhere).
"""

import logging
logger = logging.getLogger(__name__)

import math
import enum
import pathlib
import tempfile
import mimetypes
import webbrowser
import collections
from urllib.request import url2pathname
from urllib.parse import urlsplit

import gi
gi.require_version('Poppler', '0.18')
from gi.repository import Poppler

from pympress.util import fileopen


def get_extension(mime_type):
    """ Returns a valid filename extension (recognized by python) for a given mime type.

    Args:
        mime_type (`str`): The mime type for which to find an extension

    Returns:
        `str`: A file extension used for the given mimetype
    """
    if not mimetypes.inited:
        mimetypes.init()
    for ext in mimetypes.types_map:
        if mimetypes.types_map[ext] == mime_type:
            return ext


class PdfPage(enum.IntEnum):
    """ Represents the part of a PDF page that we want to draw.
    """
    #: No notes on PDF page, only falsy value
    NONE    = 0
    #: Full PDF page (without notes)
    FULL    = 1
    #: Bottom half of PDF page
    BOTTOM  = 2
    #: Top half of PDF page
    TOP     = 3
    #: Right half of PDF page
    RIGHT   = 4
    #: Left half of PDF page
    LEFT    = 5
    #: Full page + draw another page for notes, which is after the slides
    AFTER   = 6
    #: Complementary of AFTER: for a notes page, the slide page is BEFORE by half a document
    BEFORE  = 7
    #: Slides on even pages (0-indexed), notes on uneven pages
    ODD     = 8
    #: Complementary of ODD
    EVEN    = 9
    #: An arbitrary mapping of notes pages to slide pages
    MAP     = 10
    #: Reverse the arbitrary mapping MAP
    RMAP    = 11

    def complement(val):
        """ Return the enum value for the other part of the page.
        """
        return PdfPage(val ^ 1)


    def scale(val):
        """ Return the enum value that does only scaling not shifting.
        """
        return PdfPage(val | 1)


    def direction(val):
        """ Returns whether the pdf page/notes mode is horizontal or vertical.

        Returns:
            `str`: a string representing the direction that can be used as the key in the config section
        """
        if val == PdfPage.LEFT or val == PdfPage.RIGHT:
            return 'horizontal'
        elif val == PdfPage.TOP or val == PdfPage.BOTTOM:
            return 'vertical'
        elif val == PdfPage.AFTER or val == PdfPage.BEFORE:
            return 'page number'
        elif val == PdfPage.EVEN or val == PdfPage.ODD:
            return 'page parity'
        elif val == PdfPage.MAP or val == PdfPage.RMAP:
            return 'page mapping'
        else:
            return None


    def from_screen(val, x, y, x2 = None, y2 = None):
        """ Transform visible part of the page coordinates to full page coordinates.

        Pass 2 floats to transform coordinates, 4 to transform margins,
        i.e. the second pair of coordinates is taken from the opposite corner.

        Args:
            x (`float`): x coordinate on the screen, on a scale 0..1
            y (`float`): y coordinate on the screen, on a scale 0..1
            x2 (`float`): second x coordinate on the screen, from the other side, on a scale 0..1
            y2 (`float`): second y coordinate on the screen, from the other side, on a scale 0..1
        """
        if val == PdfPage.RIGHT:
            page = ((1 + x) / 2., y)
        elif val == PdfPage.LEFT:
            page = (x / 2., y)
        elif val == PdfPage.BOTTOM:
            page = (x, (1 + y) / 2.)
        elif val == PdfPage.TOP:
            page = (x, y / 2.)
        else:
            page = (x, y)

        if x2 is None or y2 is None:
            return page
        else:
            return page + val.complement().from_screen(x2, y2)


    def to_screen(val, x, y, x2 = None, y2 = None):
        """ Transform full page coordinates to visible part coordinates.

        Pass 2 floats to transform coordinates, 4 to transform margins,
        i.e. the second pair of coordinates is taken from the opposite corner.

        Args:
            x (`float`): x coordinate on the page, on a scale 0..1
            y (`float`): y coordinate on the page, on a scale 0..1
            x2 (`float`): second x coordinate on the page, from the other side, on a scale 0..1
            y2 (`float`): second y coordinate on the page, from the other side, on a scale 0..1
        """
        if val == PdfPage.RIGHT:
            screen = (x * 2 - 1, y)
        elif val == PdfPage.LEFT:
            screen = (x * 2, y)
        elif val == PdfPage.BOTTOM:
            screen = (x, y * 2 - 1)
        elif val == PdfPage.TOP:
            screen = (x, y * 2)
        else:
            screen = (x, y)

        if x2 is None or y2 is None:
            return screen
        else:
            return screen + val.complement().to_screen(x2, y2)


class Link(object):
    """ This class encapsulates one hyperlink of the document.

    Args:
        x1 (`float`):  first x coordinate of the link rectangle
        y1 (`float`):  first y coordinate of the link rectangle
        x2 (`float`):  second x coordinate of the link rectangle
        y2 (`float`):  second y coordinate of the link rectangle
        action (`function`):  action to perform when the link is clicked
    """

    #: `float`, first x coordinate of the link rectangle
    x1 = None
    #: `float`, first y coordinate of the link rectangle
    y1 = None
    #: `float`, second x coordinate of the link rectangle
    x2 = None
    #: `float`, second y coordinate of the link rectangle
    y2 = None
    #: `function`, action to be perform to follow this link
    follow = lambda *args, **kwargs: logger.error(_("no action defined for this link!"))

    def __init__(self, x1, y1, x2, y2, action):
        self.x1, self.y1, self.x2, self.y2 = x1, y1, x2, y2
        self.follow = action


    def is_over(self, x, y):
        """ Tell if the input coordinates are on the link rectangle.

        Args:
            x (`float`):  input x coordinate
            y (`float`):  input y coordinate

        Returns:
            `bool`: `True` if the input coordinates are within the link rectangle, `False` otherwise
        """
        return ((self.x1 <= x) and (x <= self.x2) and (self.y1 <= y) and (y <= self.y2))


    @staticmethod
    def build_closure(fun, *args, **kwargs):
        r""" Return a lambda that calls fun(\*args, \**kwargs), with the current value of args and kwargs.

        By creating the lambda in a new scope, we bind the arguments.

        Args:
            fun (`function`): The function to be called
            args (`tuple`): non-keyworded variable-length argument list to pass to fun()
            kwargs (`dict`): keyworded variable-length argument dict to pass to fun()
        """
        return lambda *a, **k: fun(*(tuple(args) + tuple(a)), **dict(kwargs, **k))


#: A class that holds all the properties for media files
Media = collections.namedtuple('Media', ['relative_margins', 'filename', 'autoplay', 'repeat', 'poster',
                                         'show_controls', 'type', 'start_pos', 'duration'],
                               defaults=[False, False, False, False, '', 0., 0.])


class Page(object):
    """ Class representing a single page.

    It provides several methods used by the GUI for preparing windows for
    displaying pages, managing hyperlinks, etc.

    Args:
        doc (:class:`~Poppler.Page`):  the poppler object around the page
        number (`int`):  number of the page to fetch in the document
        parent (:class:`~pympress.document.Document`):  the parent Document class
    """

    #: Page handled by this class (instance of :class:`~Poppler.Page`)
    page = None
    #: `int`, number of the current page (starting from 0)
    page_nb = -1
    #: `str` representing the page label
    page_label = None
    #: All the links in the page, as a `list` of :class:`~pympress.document.Link` instances
    links = []
    #: All the media in the page, as a `list` of :class:`~pympress.document.Media`
    medias = []
    #: `float`, page width
    pw = 0.
    #: `float`, page height
    ph = 0.
    #: All text annotations
    annotations = []
    #: Instance of :class:`~pympress.document.Document` that contains this page.
    parent = None

    def __init__(self, page, number, parent):
        self.page = page
        self.page_nb = number
        self.parent = parent
        self.links = []
        self.medias = []
        self.annotations = []

        if self.page is None:
            return

        # Get page label
        self.page_label = self.page.get_label()

        # Read page size
        self.pw, self.ph = self.page.get_size()

        # Read links on the page
        for link in self.page.get_link_mapping():
            action = self.get_link_action(link.action.type, link.action)
            my_link = Link(link.area.x1, link.area.y1, link.area.x2, link.area.y2, action)
            self.links.append(my_link)

        # Read annotations, in particular those that indicate media
        for annotation in self.page.get_annot_mapping():
            annot_type = annotation.annot.get_annot_type()

            if annot_type == Poppler.AnnotType.LINK:
                # just an Annot, not subclassed -- probably redundant with links
                continue
            elif annot_type == Poppler.AnnotType.MOVIE:
                movie = annotation.annot.get_movie()
                filepath = self.parent.get_full_path(movie.get_filename())
                if not filepath:
                    logger.error(_("Pympress can not find file ") + movie.get_filename())
                    continue

                relative_margins = Poppler.Rectangle()
                relative_margins.x1 = annotation.area.x1 / self.pw        # left
                relative_margins.x2 = 1.0 - annotation.area.x2 / self.pw  # right
                relative_margins.y1 = annotation.area.y1 / self.ph        # bottom
                relative_margins.y2 = 1.0 - annotation.area.y2 / self.ph  # top

                movie_options = {'show_controls': movie.show_controls(), 'poster': movie.need_poster()}
                try:
                    movie_options['repeat'] = movie.get_play_mode() == Poppler.MoviePlayMode.REPEAT
                    movie_options['start_pos'] = movie.get_start() / 1e9
                    movie_options['duration'] = movie.get_duration() / 1e9
                    # NB: autoplay not part of Poppler’s MovieActivationParameters struct
                except AttributeError:
                    pass  # Missing functions in pre-21.04 Poppler versions

                media = Media(relative_margins, filepath, **movie_options)
                self.medias.append(media)
                action = Link.build_closure(self.parent.play_media, hash(media))

            elif annot_type == Poppler.AnnotType.SCREEN:
                action_obj = annotation.annot.get_action()
                if not action_obj:
                    continue
                action = self.get_annot_action(action_obj.any.type, action_obj, annotation.area)
                if not action:
                    continue
            elif annot_type == Poppler.AnnotType.FILE_ATTACHMENT:
                attachment = annotation.annot.get_attachment()
                filename = pathlib.Path(attachment.name)
                with tempfile.NamedTemporaryFile('wb', suffix=filename.suffix, prefix=filename.stem, delete=False) as f:
                    # now the file name is shotgunned
                    filename = pathlib.Path(f.name)
                    self.parent.remove_on_exit(filename)
                if not attachment.save(str(filename)):
                    logger.error(_("Pympress can not extract attached file"))
                    continue
                action = Link.build_closure(fileopen, filename)
            elif annot_type in {Poppler.AnnotType.TEXT, Poppler.AnnotType.POPUP,
                                Poppler.AnnotType.FREE_TEXT}:
                # text-only annotations, hide them from screen and show them in annotations popup
                content = annotation.annot.get_contents()
                if content:
                    self.annotations.append(annotation.annot)

                self.page.remove_annot(annotation.annot)
                continue
            elif annot_type in {Poppler.AnnotType.STRIKE_OUT, Poppler.AnnotType.HIGHLIGHT,
                                Poppler.AnnotType.UNDERLINE, Poppler.AnnotType.SQUIGGLY,
                                Poppler.AnnotType.POLYGON, Poppler.AnnotType.POLY_LINE,
                                Poppler.AnnotType.SQUARE, Poppler.AnnotType.CIRCLE,
                                Poppler.AnnotType.CARET, Poppler.AnnotType.LINE,
                                Poppler.AnnotType.STAMP, Poppler.AnnotType.INK}:
                # Poppler already renders annotation of these types, nothing more can be done
                # even though the rendering isn't always perfect.
                continue
            else:
                logger.warning(_("Pympress can not interpret annotation of type:") + " {} ".format(annot_type))
                continue

            my_annotation = Link(annotation.area.x1, annotation.area.y1, annotation.area.x2, annotation.area.y2, action)
            self.links.append(my_annotation)


    def get_link_action(self, link_type, action):
        """ Get the function to be called when the link is followed.

        Args:
            link_type (:class:`~Poppler.ActionType`): The type of action to be performed
            action (:class:`~Poppler.Action`): The atcion to be performed

        Returns:
            `function`: The function to be called to follow the link
        """
        # Poppler.ActionType.RENDITION should only appear in annotations, right? Otherwise how do we know
        # where to render it? Any documentation on which action types are admissible in links vs in annots
        # is very welcome. For now, link is fallback to annot so contains all action types.
        if link_type == Poppler.ActionType.NONE:
            return lambda: None

        elif link_type == Poppler.ActionType.GOTO_DEST:
            dest_type = action.goto_dest.dest.type
            if dest_type == Poppler.DestType.NAMED:
                dest = self.parent.doc.find_dest(action.goto_dest.dest.named_dest)
                if dest:
                    return Link.build_closure(self.parent.goto_page, dest.page_num - 1)
                else:
                    warning = _('Unrecognized named destination: ') + str(action.goto_dest.dest.named_dest)
            elif dest_type != Poppler.DestType.UNKNOWN:
                return Link.build_closure(self.parent.goto_page, action.goto_dest.dest.page_num - 1)

        elif link_type == Poppler.ActionType.NAMED:
            dest_name = action.named.named_dest
            dest = self.parent.doc.find_dest(dest_name)

            if dest:
                return Link.build_closure(self.parent.goto_page, dest.page_num)
            elif dest_name == "GoBack":
                return self.parent.goto_prev_hist
            elif dest_name == "GoForward":
                return self.parent.goto_next_hist
            elif dest_name == "FirstPage":
                return Link.build_closure(self.parent.goto_page, 0)
            elif dest_name == "PrevPage":
                return Link.build_closure(self.parent.goto_page, self.page_nb - 1)
            elif dest_name == "NextPage":
                return Link.build_closure(self.parent.goto_page, self.page_nb + 1)
            elif dest_name == "LastPage":
                return Link.build_closure(self.parent.goto_page, self.parent.pages_number() - 1)
            elif dest_name == "GoToPage":
                # Same as the 'G' action which allows one to pick a page to jump to
                return Link.build_closure(self.parent.start_editing_page_number, )
            elif dest_name == "Find":
                # TODO popup a text box and search results with Page.find_text
                # http://lazka.github.io/pgi-docs/Poppler-0.18/classes/Page.html#Poppler.Page.find_text
                warning = _("Pympress does not yet support link type \"{}\" to \"{}\"").format(link_type, dest_name)
            else:
                # TODO find out other possible named actions?
                warning = _("Pympress does not recognize link type \"{}\" to \"{}\"").format(link_type, dest_name)

        elif link_type == Poppler.ActionType.LAUNCH:
            launch = action.launch
            if launch.params:
                logger.warning("ignoring params: " + str(launch.params))

            filepath = self.parent.get_full_path(launch.file_name)
            if not filepath:
                logger.error("can not find file " + launch.file_name)
                return lambda: None

            else:
                return Link.build_closure(fileopen, filepath)

        elif link_type == Poppler.ActionType.URI:
            return Link.build_closure(webbrowser.open_new_tab, action.uri.uri)

        elif link_type == Poppler.ActionType.RENDITION:  # Poppler 0.22
            warning = _("Pympress does not yet support link type \"{}\"").format(link_type)
        elif link_type == Poppler.ActionType.MOVIE:  # Poppler 0.20
            warning = _("Pympress does not yet support link type \"{}\"").format(link_type)
        elif link_type == Poppler.ActionType.GOTO_REMOTE:
            warning = _("Pympress does not yet support link type \"{}\"").format(link_type)
        elif link_type == Poppler.ActionType.OCG_STATE:
            warning = _("Pympress does not yet support link type \"{}\"").format(link_type)
        elif link_type == Poppler.ActionType.JAVASCRIPT:
            warning = _("Pympress does not yet support link type \"{}\"").format(link_type)
        elif link_type == Poppler.ActionType.UNKNOWN:
            warning = _("Pympress does not yet support link type \"{}\"").format(link_type)
        else:
            warning = _("Pympress does not recognize link type \"{}\"").format(link_type)

        logger.info(warning)
        return Link.build_closure(logger.warning, _('Unsupported link clicked. ') + warning)


    def get_annot_action(self, link_type, action, rect):
        """ Get the function to be called when the link is followed.

        Args:
            link_type (:class:`~Poppler.ActionType`): The link type
            action (:class:`~Poppler.Action`): The action to be performed when the link is clicked
            rect (:class:`~Poppler.Rectangle`): The region of the page where the link is

        Returns:
            `function`: The function to be called to follow the link
        """
        if link_type == Poppler.ActionType.RENDITION:
            media = action.rendition.media
            if media.is_embedded():
                ext = get_extension(media.get_mime_type())
                with tempfile.NamedTemporaryFile('wb', suffix=ext, prefix='pdf_embed_', delete=False) as f:
                    # now the file name is shotgunned
                    filename = pathlib.Path(f.name)
                    self.parent.remove_on_exit(filename)
                if not media.save(str(filename)):
                    logger.error(_("Pympress can not extract embedded media"))
                    return None
            else:
                filename = self.parent.get_full_path(media.get_filename())
                if not filename:
                    logger.error(_("Pympress can not find file ") + media.get_filename())
                    return None

            relative_margins = Poppler.Rectangle()
            relative_margins.x1 = rect.x1 / self.pw        # left
            relative_margins.x2 = 1.0 - rect.x2 / self.pw  # right
            relative_margins.y1 = rect.y1 / self.ph        # bottom
            relative_margins.y2 = 1.0 - rect.y2 / self.ph  # top

            media_options = {'type': media.get_mime_type()}
            try:
                media_options['autoplay'] = media.get_auto_play()
                media_options['show_controls'] = media.get_show_controls()
                media_options['repeat'] = media.get_repeat_count() - 1
                # NB: no poster in Poppler’s MediaParameters
            except AttributeError:
                pass
            media = Media(relative_margins, filename, **media_options)
            self.medias.append(media)
            return Link.build_closure(self.parent.play_media, hash(media))

        else:
            return self.get_link_action(link_type, action)


    def number(self):
        """ Get the page number.
        """
        return self.page_nb


    def label(self):
        """ Get the page label.
        """
        return self.page_label


    def get_link_at(self, x, y, dtype=PdfPage.FULL):
        """ Get the :class:`~pympress.document.Link` corresponding to the given position.

        Returns `None` if there is no link at this position.

        Args:
            x (`float`):  horizontal coordinate
            y (`float`):  vertical coordinate
            dtype (:class:`~pympress.document.PdfPage`):  the type of document to consider

        Returns:
            :class:`~pympress.document.Link`: the link at the given coordinates
            if one exists, `None` otherwise
        """
        x, y = dtype.from_screen(x, y)

        xx = self.pw * x
        yy = self.ph * (1. - y)

        for link in self.links:
            if link.is_over(xx, yy):
                return link

        return None


    def get_size(self, dtype=PdfPage.FULL):
        """ Get the page size.

        Args:
            dtype (:class:`~pympress.document.PdfPage`):  the type of document to consider

        Returns:
            `(float, float)`: page size
        """
        return dtype.scale().from_screen(self.pw, self.ph)


    def get_aspect_ratio(self, dtype=PdfPage.FULL):
        """ Get the page aspect ratio.

        Args:
            dtype (:class:`~pympress.document.PdfPage`):  the type of document to consider

        Returns:
            `float`: page aspect ratio
        """
        w, h = self.get_size(dtype)
        return w / h


    def get_annotations(self):
        """ Get the list of text annotations on this page.

        Returns:
            `list` of `str`: annotations on this page
        """
        return self.annotations


    def new_annotation(self, pos, rect=None):
        """ Add an annotation to this page

        Args:
            pos (`int`): The position in the list of annotations in which to insert this annotation
            rect (:class:`~Poppler.Rectangle`): A rectangle for the position of this annotation

        Returns:
            :class:`~Poppler.Annot`: A new annotation on this page
        """
        if pos < 0:
            pos = 0
        if pos > len(self.annotations):
            pos = len(self.annotations)

        if rect is None:
            rect = Poppler.Rectangle()
            rect.x1 = self.pw - 20
            rect.x2 = rect.x1 + 20
            rect.y2 = self.ph - len(self.annotations) * 20
            rect.y1 = rect.y2 - 20

        new_annot = Poppler.AnnotText.new(self.parent.doc, rect)
        new_annot.set_icon(Poppler.ANNOT_TEXT_ICON_NOTE)
        self.annotations.insert(pos, new_annot)
        self.parent.made_changes()
        return new_annot


    def set_annotation(self, pos, value):
        """ Add an annotation to this page

        Args:
            pos (`int`): The number of the annotation
            value (`str`): The new contents of the annotation
        """
        rect = self.annotations[pos].get_rectangle()
        self.remove_annotation(pos)
        self.new_annotation(pos, rect).set_contents(value)


    def remove_annotation(self, pos):
        """ Add an annotation to this page

        Args:
            pos (`int`): The number of the annotation
        """
        self.parent.made_changes()
        del self.annotations[pos]


    def get_media(self):
        """ Get the list of medias this page might want to play.

        Returns:
            `list`: medias in this page
        """
        return self.medias


    def render_cairo(self, cr, ww, wh, dtype=PdfPage.FULL):
        """ Render the page on a Cairo surface.

        Args:
            cr (:class:`~Gdk.CairoContext`):  target surface
            ww (`int`):  target width in pixels
            wh (`int`):  target height in pixels
            dtype (:class:`~pympress.document.PdfPage`):  the type of document that should be rendered
        """
        pw, ph = self.get_size(dtype)

        cr.set_source_rgb(1, 1, 1)

        # Scale
        scale = min(ww / pw, wh / ph)
        cr.scale(scale, scale)

        cr.rectangle(0, 0, pw, ph)
        cr.fill()

        # For "regular" pages, there is no problem: just render them.
        # For other pages (i.e. half of a page), the widget already has correct
        # dimensions so we don't need to deal with that. But for right and bottom
        # halfs we must translate the output in order to only show the correct half.
        if dtype == PdfPage.RIGHT:
            cr.translate(-pw, 0)
        elif dtype == PdfPage.BOTTOM:
            cr.translate(0, -ph)

        self.page.render(cr)


    def can_render(self):
        """ Informs that rendering *is* necessary (avoids checking the type).

        Returns:
            `bool`: `True`, do rendering
        """
        return True



class Document(object):
    """ This is the main document handling class.

    The page numbering starts as 0 and is aware of notes (i.e. number of pages may change to account for note pages).
    The document page numbers are the same as in Poppler, and also start at 0 but do not depend on notes.

    Args:
        builder (:class:`pympress.builder.Builder`):  A builder to load callbacks
        pop_doc (:class:`~pympress.Poppler.Document`):  Instance of the Poppler document that this class will wrap
        uri (`str`):  URI of the PDF file to open
        page (`int`):  page number to which the file should be opened
    """

    #: Current PDF document (:class:`~Poppler.Document` instance)
    doc = None
    #: `str` full path to pdf
    uri = None
    #: :class:`~pathlib.Path` to pdf if uri is a file: URI
    path = None
    #: Number of pages in the document
    nb_pages = -1
    #: Pages cache (`dict` of :class:`~pympress.document.Page`). This makes
    #: navigation in the document faster by avoiding calls to Poppler when loading
    #: a page that has already been loaded.
    pages_cache = {}
    #: `set` of :class:`~pathlib.Path` representing the temporary files which need to be removed
    temp_files = set()
    #: History of pages we have visited, using note-aware page numbers
    history = []
    #: Our position in the history
    hist_pos = -1
    #: `list` of slide page labels, indexed on note-aware page numbers
    page_labels = []
    #: `list` of all the page labels, indexed on document page numbers
    doc_page_labels = []
    #: `list` of (slide's document page number, notes' document page number) tuples, or `None` if there are no notes
    notes_mapping = None
    #: `bool` indicating whether there were modifications to the document
    changes = False

    #: callback, to be connected to :func:`~pympress.extras.Media.play`
    play_media = lambda *args: None
    #: callback, to be connected to :func:`~pympress.editable_label.PageNumber.start_editing`
    start_editing_page_number = lambda *args: None
    #: callback, to be connected to :func:`~pympress.ui.UI.goto_page`
    navigate = lambda *args: None

    def __init__(self, builder, pop_doc, uri):
        if builder is not None:
            # Connect callbacks
            self.play_media                = builder.get_callback_handler('medias.play')
            self.start_editing_page_number = builder.get_callback_handler('page_number.start_editing')
            self.goto_page                 = builder.get_callback_handler('goto_page')
            self.goto_next_hist            = builder.get_callback_handler('doc_hist_next')
            self.goto_prev_hist            = builder.get_callback_handler('doc_hist_prev')

        # Setup PDF file
        self.uri = uri
        self.doc = pop_doc
        self.changes = False

        if uri is not None:
            uri_parts = urlsplit(uri, scheme='file')
            self.path = pathlib.Path(url2pathname(uri_parts.path))
            if uri_parts.scheme == 'file':
                self.path = pathlib.Path.cwd().joinpath(self.path.name)
        else:
            self.path = None

        # Pages numbers and labels
        self.nb_pages = 0 if pop_doc is None else self.doc.get_n_pages()
        self.doc_page_labels = [self.doc.get_page(n).get_label() for n in range(self.nb_pages)]
        self.page_labels = self.doc_page_labels

        # Pages cache
        self.pages_cache = {}


    def get_structure(self, index_iter = None):
        """ Gets the structure of the document from its index.

        Recursive, pass the iterator.

        Args:
            index_iter (:class:`~Poppler.IndexIter` or `None`): the iterator for the child index to explore.

        Returns:
            `list`: A list of tuples (depth, page number, title)
        """
        try:
            if index_iter is None:
                index_iter = Poppler.IndexIter(self.doc)
        except TypeError:
            return {}
        if index_iter is None:
            return {}

        index = {}
        while True:
            action = index_iter.get_action()
            title = ''
            page = None
            try:
                if action.type == Poppler.ActionType.GOTO_DEST:
                    title = action.goto_dest.title
                    if action.goto_dest.dest.type == Poppler.DestType.NAMED:
                        page = self.doc.find_dest(action.goto_dest.dest.named_dest).page_num - 1
                    elif action.goto_dest.dest.type == Poppler.DestType.UNKNOWN:
                        raise AssertionError('Unknown type of destination')
                    else:
                        page = action.goto_dest.dest.page_num - 1
                else:
                    raise AssertionError('Unexpected type of action')

            except Exception:
                logger.error(_('Unexpected action in index "{}"').format(action.type))
                page = None

            new_entry = {'title': title}
            child = index_iter.get_child()
            if child:
                new_entry['children'] = self.get_structure(child)
                if page is None:
                    page = min(new_entry['children'])

            # there should not be synonymous sections, correct the page here to a better guess
            if page in index:
                lower_bound = max(index.keys())
                find = index[lower_bound]
                while 'children' in find:
                    lower_bound = max(find['children'].keys())
                    find = find['children'][lower_bound]

                try:
                    page = min(number for number, label in enumerate(self.page_labels)
                               if label == self.page_labels[page] and number > lower_bound)
                except ValueError:  # empty iterator
                    page = lower_bound + 1


            if page is not None:
                index[page] = new_entry

            if not index_iter.next():
                break

        return index


    @staticmethod
    def create(builder, uri):
        """ Initializes a Document by passing it a :class:`~Poppler.Document`.

        Args:
            builder (:class:`pympress.builder.Builder`):  A builder to load callbacks
            uri (`str`):  URI to the PDF file to open
            page (`int`):  page number to which the file should be opened

        Returns:
            :class:`~pympress.document.Document`: The initialized document
        """
        if uri is None:
            doc = EmptyDocument()
        else:
            poppler_doc = Poppler.Document.new_from_file(uri, None)
            doc = Document(builder, poppler_doc, uri)

        return doc


    def made_changes(self):
        """ Notify the document that some changes were made (e.g. annotations edited)
        """
        self.changes = True


    def has_changes(self):
        """ Return whether that some changes were made (e.g. annotations edited)
        """
        return self.changes


    def save_changes(self, dest_uri=None):
        """ Save the changes

        Args:
            dest_uri (`str` or `None`): The URI where to save the file, or None to save in-place
        """
        if self.doc is None:
            return

        for page in self.pages_cache.values():
            for annot in page.get_annotations():
                page.page.add_annot(annot)

        if dest_uri is not None and dest_uri != self.uri:
            if self.doc.save(dest_uri):
                self.changes = False
        else:
            # We can’t overwrite the current file directly, so create a temporary file and then overwrite
            with tempfile.NamedTemporaryFile('wb', suffix=self.path.suffix, prefix=self.path.stem, dir=self.path.parent,
                                             delete=False) as f:
                temp_path = pathlib.Path(f.name)

            if self.doc.save(temp_path.as_uri()):
                self.changes = False
                temp_path.replace(self.path)

        for page in self.pages_cache.values():
            for annot in page.get_annotations():
                page.page.remove_annot(annot)


    def guess_notes(self, horizontal, vertical, current_page=0):
        """ Get our best guess for the document mode.

        Args:
            horizontal (`str`): A string representing the preference for horizontal slides
            vertical (`str`): A string representing the preference for vertical slides

        Returns:
            :class:`~pympress.document.PdfPage`: the notes mode
        """
        if any(label.startswith('notes:') for label in self.page_labels):
            return PdfPage.MAP

        page = self.page(current_page)
        if page is None or not page.can_render():
            return PdfPage.NONE

        ar = page.get_aspect_ratio()

        # Check whether we have N slides with one aspect ratio then N slides with a different aspect ratio
        # that is the sign if Libreoffice notes pages
        if self.nb_pages and self.nb_pages % 2 == 0:
            half_doc = self.nb_pages // 2
            ar_slides = self.page(0).get_aspect_ratio()
            ar_notes = self.page(half_doc).get_aspect_ratio()
            if ar_slides != ar_notes and \
                    all(self.page(p).get_aspect_ratio() == ar_slides for p in range(1, half_doc)) and \
                    all(self.page(half_doc + p).get_aspect_ratio() == ar_notes for p in range(1, half_doc)):
                return PdfPage.AFTER

        # "Regular" slides will have an aspect ratio of 4/3, 16/9, 16/10... i.e. in the range [1..2]
        # So if the aspect ratio is >= 2, we can assume it is a document with notes on the side.
        if ar >= 2:
            try:
                return PdfPage[horizontal.upper()]
            except KeyError:
                return PdfPage.RIGHT

        # Make exception for classic american letter format and ISO (A4, B5, etc.)
        if abs(ar - 8.5 / 11) < 1e-3 or abs(ar - 1 / math.sqrt(2)) < 1e-3:
            return PdfPage.NONE

        # If the aspect ratio is < 1, we can assume it is a document with notes above or below.
        if ar < 1:
            try:
                return PdfPage[vertical.upper()]
            except KeyError:
                return PdfPage.BOTTOM

        return PdfPage.NONE


    def set_notes_pos(self, notes_direction):
        """ Set whether where the notes pages are relative to normal pages

        Valid values are returned by :meth:`~pympress.document.PdfPage.direction`
        - page number (aka Libreoffice notes mode)
        - page parity (can not be detected automatically, where every other page contains notes)
        - page mapping (where labels of notes pages are corresponding slide labels prefixed with “notes:”)

        Args:
            notes_direction (`str`):  Where the notes pages are
        """
        if notes_direction == 'page number':
            self.notes_mapping = [(n, n + self.nb_pages // 2) for n in range(self.nb_pages // 2)]
        elif notes_direction == 'page parity':
            self.notes_mapping = [(n, n + 1) for n in range(0, self.nb_pages, 2)]
        elif notes_direction == 'page mapping':
            notes_mapping = collections.OrderedDict()
            for n, (label, prev_label) in enumerate(zip(self.doc_page_labels, [None, *self.doc_page_labels[:-1]])):
                # Here the condition (could be adjusted) is 2 successive pages labeled <label> and notes:<label>,
                # with the prior page not being a notes page.
                if n - 1 in notes_mapping and label == 'notes:' + prev_label:
                    notes_mapping[n - 1] = n
                else:
                    notes_mapping[n] = None
            self.notes_mapping = list(notes_mapping.items())
        else:
            self.notes_mapping = None
            self.page_labels = self.doc_page_labels
            return

        self.page_labels = [self.doc_page_labels[page] for page, note in self.notes_mapping]


    def page(self, number):
        """ Get the specified page.

        Args:
            number (`int`):  number of the page to return

        Returns:
            :class:`~pympress.document.Page`: the wanted page, or `None` if it does not exist
        """
        if number >= self.pages_number() or number < 0:
            return None

        if self.notes_mapping is not None:
            number = self.notes_mapping[number][0]
            if number < 0:
                return None

        if number not in self.pages_cache:
            self.pages_cache[number] = Page(self.doc.get_page(number), number, self)
        return self.pages_cache[number]


    def notes_page(self, number):
        """ Get the specified page.

        Args:
            number (`int`):  number of the page to return

        Returns:
            :class:`~pympress.document.Page`: the wanted page, or `None` if it does not exist
        """
        if number >= self.pages_number() or number < 0:
            return None

        if self.notes_mapping is not None:
            number = self.notes_mapping[number][1]

        if number is None:
            return None

        if number not in self.pages_cache:
            self.pages_cache[number] = Page(self.doc.get_page(number), number, self)
        return self.pages_cache[number]


    def pages_number(self):
        """ Get the number of pages in the document.

        Returns:
            `int`: the number of pages in the document
        """
        return len(self.notes_mapping) if self.notes_mapping is not None else self.nb_pages


    def has_labels(self):
        """ Return whether this document has useful labels.

        Returns:
            `bool`: False iff there are no labels or they are just the page numbers
        """
        return self.doc_page_labels != [str(n + 1) for n in range(self.nb_pages)]


    def get_last_label_pages(self):
        """ Return the last page number for each consecutively distinct page label

        In other words, squash together consecutive same labels
        """
        last = None
        pages = []
        for page, label in enumerate(self.page_labels):
            if label != last:
                pages.append(page)
            last = label
        return pages


    def lookup_label(self, label, prefix_unique=True):
        """ Find a page from its label.

        Args:
            label (`str`): the label we are searching for
            prefix_unique (`bool`): whether a prefix match should be unique, e.g. when the user is still typing

        Returns:
            `int`: the page
        """
        # somehow this always returns None:
        # page = self.doc.get_page_by_label(label).get_index()

        # make a shortlist: squash synonymous labels, keeping the last one
        compatible_labels = {l: n for n, l in enumerate(self.page_labels) if l.lower().startswith(label.lower())}

        if len(compatible_labels) == 1:
            return set(compatible_labels.values()).pop()

        # try exact match
        try:
            return compatible_labels[label]
        except KeyError:
            pass

        # try case-insensitive match, prefix case-sensitive match, prefix case-insensitive match (unless prefix_unique)
        filters = [lambda lbl: len(lbl) == len(label), lambda lbl: lbl.startswith(label), lambda lbl: not prefix_unique]
        for filtering in filters:
            try:
                found = next(label for label in compatible_labels if filtering(label))
            except StopIteration:
                pass
            else:
                return compatible_labels[found]

        return None


    def goto(self, number):
        """ Switch to another page. Validates the number and returns one in the correct range. Also updates history.

        Args:
            number (`int`):  number of the destination page
        """
        if number < 0:
            number = 0
        if number >= self.pages_number():
            number = self.pages_number() - 1

        if 0 <= self.hist_pos < len(self.history) and self.history[self.hist_pos] == number:
            return number

        # chop off history where we were and append
        self.hist_pos = min(len(self.history), self.hist_pos + 1)
        del self.history[self.hist_pos:]
        self.history.append(number)

        return number


    def label_after(self, page):
        """ Switch to the next page with different label.

        If we're within a set of pages with the same label we want to go to the last one.
        """
        labels_after = enumerate(self.page_labels[page + 1:], page + 1)

        try:
            next_page, next_label = next(labels_after)
        except StopIteration:
            # we're already at the last page!
            return page

        # will stop as soon as next_page + 1 (aka following_page) is a different label or due to end of iterator
        for following_page, following_label in labels_after:
            if following_label == next_label:
                next_page = following_page
            else:
                break

        return next_page


    def label_before(self, page):
        """ Switch to the previous page with different label.

        If we're within a set of pages with the same label we want to go *before* the first one.
        """
        # will stop as soon as we find a different label or due to end of iterator
        for prev_page, prev_label in enumerate(reversed(self.page_labels[:page])):
            if prev_label != self.page_labels[page]:
                return page - 1 - prev_page

        return 0


    def hist_next(self, *args):
        """ Switch to the page we viewed next.
        """
        if self.hist_pos + 1 == len(self.history):
            return None

        self.hist_pos += 1
        return self.history[self.hist_pos]


    def hist_prev(self, *args):
        """ Switch to the page we viewed before.
        """
        if self.hist_pos == 0:
            return None

        self.hist_pos -= 1
        return self.history[self.hist_pos]


    def get_uri(self):
        """ Gives access to the URI, rather than the path, of this document.

        Returns:
            `str`: the URI to the file currently opened.
        """
        return self.uri


    def get_full_path(self, filename):
        """ Returns full path, extrapolated from a path relative to this document or to the current directory.

        Args:
            filename (:class:`~pathlib.Path` or `str`):  Name of the file or relative path to it

        Returns:
            :class:`~pathlib.Path`: the full path to the file or None if it doesn't exist
        """
        filename = pathlib.Path(filename)
        if filename.is_absolute():
            return filename.resolve() if filename.exists() else None

        for dirname in [pathlib.Path(url2pathname(urlsplit(self.uri).path)).parent, pathlib.Path.cwd()]:
            filepath = dirname.joinpath(filename)
            if filepath.exists():
                return filepath


    def remove_on_exit(self, filename):
        """ Remember a temporary file to delete later.

        Args:
            filename (:class:`~pathlib.Path`): The path to the file to delete
        """
        self.temp_files.add(filename)


    def cleanup_media_files(self):
        """ Removes all files that were extracted from the pdf into the filesystem.
        """
        for f in self.temp_files:
            if f.exists():
                f.unlink()
        self.temp_files.clear()


class EmptyPage(Page):
    """ A dummy page, placeholder for when there are no valid pages around.

    This page is a non-notes page with an aspect ratio of 1.3 and nothing else inside.
    Also, it has no "rendering" capability, and is made harmless by overriding its render function.
    """

    def __init__(self):
        super(EmptyPage, self).__init__(None, -1, None)
        self.page_label = None
        # by default, anything that will have a 1.3 asapect ratio
        self.pw, self.ph = 1.3, 1.0


    def render_cairo(self, cr, ww, wh, dtype=PdfPage.FULL):
        """ Overriding this purely for safety: make sure we do not accidentally try to render.

        Args:
            cr (:class:`~Gdk.CairoContext`):  target surface
            ww (`int`):  target width in pixels
            wh (`int`):  target height in pixels
            dtype (:class:`~pympress.document.PdfPage`):  the type of document that should be rendered
        """
        pass


    def can_render(self):
        """ Informs that rendering is *not* necessary (avoids checking the type).

        Returns:
            `bool`: `False`, no rendering
        """
        return False


class EmptyDocument(Document):
    """ A dummy document, placeholder for when no document is open.
    """
    def __init__(self):
        super(EmptyDocument, self).__init__(None, None, None)
        self.pages_cache[-1] = EmptyPage()


    def page(self, number):
        """ Retrieve a page from the document.

        Args:
            number (`int`): page number to be retrieved

        Returns:
            :class:`~pympress.document.EmptyPage` or `None`: -1 returns the empty page so we can display something.
        """
        return self.pages_cache[number] if number in self.pages_cache else None


    def notes_page(self, number):
        """ Retrieve a page from the document.

        Args:
            number (`int`): page number to be retrieved

        Returns:
            :class:`~pympress.document.EmptyPage` or `None`: -1 returns the empty page so we can display something.
        """
        return self.pages_cache[number] if number in self.pages_cache else None


##
# Local Variables:
# mode: python
# indent-tabs-mode: nil
# py-indent-offset: 4
# fill-column: 80
# end:
